/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamdispatch.h"
#include "gdamdevice.h"
#include <gdam/gdam-protocol.h>
/*
 * Code which handles the global communication.
 *
 * In particular the gdam_source_id/gdam_device_id interface
 * is accessible here.
 */
static GHashTable* source_map     = NULL;
static GHashTable* sink_map       = NULL;
static int         last_source_id = 0;
static int         last_sink_id   = 0;
extern GHashTable* gdam_dispatch_methods;

GdamSource* gdam_dispatch_get_source(int gdam_source_id)
{
	gpointer rv;
	if (!source_map) return NULL;
	rv = g_hash_table_lookup(source_map, GINT_TO_POINTER(gdam_source_id));
	if (!rv) return NULL;
	gtk_object_ref((GtkObject*)rv);
	return (GdamSource*)rv;
}

GdamSink* gdam_dispatch_get_sink(int gdam_sink_id)
{
	gpointer rv;
	if (!sink_map) return NULL;
	rv = g_hash_table_lookup(sink_map, GINT_TO_POINTER(gdam_sink_id));
	if (!rv) return NULL;
	gtk_object_ref((GtkObject*)rv);
	return (GdamSink*)rv;
}



GdamBuffer* gdam_dispatch_buffer(GdamConnection* connection,
                                 GdamPermissions* perms,
				 GdamBuffer* command)
{
	int type;
	GdamBuffer* rv = gdam_buffer_new();
	GdamDispatchMessageFunc dispatcher;

	if (!gdam_buffer_parse_int(command, &type)) {
		g_warning("error parsing command type from header");
		return NULL;
	}
#if DEBUG_DISPATCHING
	gdam_debug("gdam_dispatch: message of type %d", type);
#endif
	dispatcher = gdam_dispatch_message_find_by_type(type);
	if (!dispatcher) {
		if (GDAM_MESSAGE_FIRST_MESSAGE < type
		 && type < GDAM_MESSAGE_FIRST_RESPONSE) {
			g_warning("message was in valid range..."
				  "presumably a unimplemented command");
			gdam_buffer_append_int(rv, GDAM_RESPONSE_FAILED);
			gdam_buffer_append_int(rv, type);
			{
				int tmp;
				g_assert(gdam_buffer_peek_int(rv, &tmp));
				g_assert(tmp == GDAM_RESPONSE_FAILED);
				g_assert(rv->size == 8);
			}
			return rv;
		} else if (GDAM_MESSAGE_FIRST_RESPONSE <= type
		 && type < GDAM_MESSAGE_LAST_MESSAGE) {
		 	g_warning("got a response from the client... "
				  "protocol error");
		} else if (type == GDAM_MESSAGE_FIRST_MESSAGE) {
			g_warning("GDAM_MESSAGE_FIRST_MESSAGE "
				  "should never be a message type");
		} else if (type == GDAM_MESSAGE_LAST_MESSAGE) {
			g_warning("GDAM_MESSAGE_LAST_MESSAGE "
				  "should never be a message type");
		} else {
			g_warning("client message was completely "
				  "invalid-- type id out of range");
		}
		gdam_buffer_destroy(rv);
		return NULL;
	}

	/* Strange as it may seem, we don't pass whether the
	 * command succeeded or failed to the caller.
	 * Maybe we should... */
	switch ( (*dispatcher)(connection, perms, command, rv) ) {
		case GDAM_DISPATCH_INDICATE_SUCCESS:
			gdam_buffer_append_int(rv, GDAM_RESPONSE_SUCCESS);
			gdam_buffer_append_int(rv, type);
		case GDAM_DISPATCH_SUCCESS:
			break;

		case GDAM_DISPATCH_TOO_SHORT:
			g_warning("message of type `%s' was too short",
				gdam_protocol_get_enum_name("message", type));
		case GDAM_DISPATCH_INDICATE_FAIL:
			gdam_buffer_append_int(rv, GDAM_RESPONSE_FAILED);
			gdam_buffer_append_int(rv, type);
		case GDAM_DISPATCH_FAIL:
			break;

		default:
			g_assert_not_reached();
	}
	return rv;
}

int gdam_dispatch_obtain_source_id(GdamSource* source)
{
	if (last_source_id++ == 0) {
		source_map = g_hash_table_new(NULL, NULL);
	}
	g_hash_table_insert(source_map,
	                    GINT_TO_POINTER(last_source_id),
			    source);
	return last_source_id;
}

int gdam_dispatch_obtain_sink_id(GdamSink* sink)
{
	if (last_sink_id++ == 0) {
		sink_map = g_hash_table_new(NULL, NULL);
	}
	g_hash_table_insert(sink_map,
	                    GINT_TO_POINTER(last_sink_id),
			    sink);
	return last_sink_id;
}

void gdam_dispatch_release_source_id(int source_id)
{
	g_hash_table_remove(source_map, GINT_TO_POINTER(source_id));
}

void gdam_dispatch_release_sink_id(int sink_id)
{
	g_hash_table_remove(sink_map, GINT_TO_POINTER(sink_id));
}
