/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsource.h"
#include "gdamexternal.h"
#include "gdamcleanup.h"
#include <gdam/gdambuffer.h>
#include <gdam/gdam-protocol.h>
#include "gdamconnection.h"
#include "gdamevent.h"
#include "gdamserverdebug.h"
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>

/* Wait until 1/8 the maximum buffer is full,
 * or MAX_TO_WAIT_FOR samples are ready
 * before saying the source is ready to read. */
#define FRACTION_TO_WAIT_FOR	4
#define MAX_TO_WAIT_FOR		8192

/*
 * Notes:
 *   - there is no way to run an external source generally from
 *     the server.  That would be a big security hole.
 */

static int         gdam_external_read      (GdamSource*	external,
                                       	    GdamSample*	buf,
				       	    int		n);

static void        gdam_external_class_init       (GdamSourceClass* klass);
static void        gdam_external_init             (GdamExternal*    external);
static void        gdam_external_destroy          (GdamExternal*    source);

static GdamTime    gdam_external_get_position     (GdamSource*      source);
static gboolean    gdam_external_is_ready_to_read (GdamSource*      source);
static int	   gdam_external_read             (GdamSource*      input,
					           GdamSample*      samples,
					           int              max_read);
static void        gdam_external_miss_data        (GdamSource*      input,
					           int              lost);
static gboolean    gdam_external_prepare          (gpointer         src_data,
                                                   GTimeVal*        curtime,
                                                   gint*            timeout,
		                                   gpointer         user_data);
static gboolean    gdam_external_check            (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data);
static gboolean    gdam_external_dispatch         (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data);

static inline void add_poll_fd(GdamExternal* external)
{
	g_main_add_poll(&external->poll_fd, G_PRIORITY_DEFAULT);
	external->has_poll_fd = TRUE;
}
static inline void remove_poll_fd(GdamExternal* external)
{
	g_main_remove_poll(&external->poll_fd);
	external->has_poll_fd = FALSE;
}


	/** derivers might usefully define these,
	 ** but we didn't want to give the client end
	 ** the general controls until they stabilize....
	 ** nonetheless, a generic `pause' and `resume'
	 ** and so on might be later implemented here. */
#if 0
static void        gdam_external_set_arg   (GtkObject*        external,
                                            GtkArg*           arg,
					    guint             arg_id);
static void        gdam_external_get_arg   (GtkObject*        external,
                                            GtkArg*           arg,
					    guint             arg_id);
#endif

static GtkObjectClass*  parent_class = NULL;

GtkType gdam_external_get_type()
{
	static GtkType external_type = 0;
	if (!external_type) {
		static const GtkTypeInfo external_info = {
			"GdamExternal",
			sizeof(GdamExternal),
			sizeof(GdamExternalClass),
			(GtkClassInitFunc) gdam_external_class_init,
			(GtkObjectInitFunc) gdam_external_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		external_type = gtk_type_unique(GDAM_TYPE_SOURCE,
		                                &external_info);
	}
	return external_type;
}
static void gdam_external_set_length(GtkObject* object, GdamArg* arg)
{
	GdamExternal* external = GDAM_EXTERNAL(object);
	external->samples_remaining = GDAM_ARG_INT64(arg);
}
static void gdam_external_get_length(GtkObject* object, GdamArg* arg)
{
	GdamExternal* external = GDAM_EXTERNAL(object);
	GDAM_ARG_INT64(arg) = external->samples_remaining;
}

static void gdam_external_set_buffer_size(GtkObject* object, GdamArg* arg)
{
	GdamExternal* external = GDAM_EXTERNAL(object);
	external->max_buffer_size = GDAM_ARG_INT(arg);
}
static void gdam_external_get_buffer_size(GtkObject* object, GdamArg* arg)
{
	GdamExternal* external = GDAM_EXTERNAL(object);
	GDAM_ARG_INT(arg) = external->max_buffer_size;
}

static void gdam_external_real_do_exec(GdamExternal* external,
                                       char**        command)
{
	execvp(command[0], command);
}

static void gdam_external_class_init(GdamSourceClass* source_class)
{
	GtkObjectClass* object_class = (GtkObjectClass*)source_class;
	GdamExternalClass* external_class = GDAM_EXTERNAL_CLASS(source_class);
	if (!parent_class)
		parent_class = gtk_type_class(GDAM_TYPE_SOURCE);
	source_class->read = gdam_external_read;
	source_class->miss_data = gdam_external_miss_data;
	source_class->is_ready_to_read = gdam_external_is_ready_to_read;
	source_class->get_position = gdam_external_get_position;
	external_class->do_exec = gdam_external_real_do_exec;
	object_class->destroy = (void(*)(GtkObject*)) gdam_external_destroy;

	gdam_arg_register(object_class,
	                  "length",
			  GDAM_ARG_TYPE_INT64,
			  gdam_external_set_length,
			  gdam_external_get_length,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "buffer_size",
			  GDAM_ARG_TYPE_INT,
			  gdam_external_set_buffer_size,
			  gdam_external_get_buffer_size,
			  GDAM_ARG_VITAL);
}

static int default_buffer_size = 64 * 1024;

int gdam_external_get_default_buffer_size()
{
	return default_buffer_size;
}
void gdam_external_set_default_buffer_size(int buffer_size)
{
	default_buffer_size = buffer_size;
}

static void gdam_external_init(GdamExternal* external)
{
	external->process_id      = -1;
	external->poll_fd.fd      = -1;
	external->poll_fd.events  = G_IO_IN|G_IO_ERR|G_IO_HUP;
	external->reader_source   = 0;
	external->skipped         = 0;
	external->max_buffer_size = gdam_external_get_default_buffer_size();
	external->samples_remaining = -1;
	gdam_buffer_construct(&external->buffer);
}

/*=====================================================================*
 *                      gdam_external_kill_command                     *
 *                                                                     *
 * external                                                            *
 * the source which should stop running the command.                   *
 *=====================================================================*/
gboolean gdam_external_kill_command(GdamExternal* external,
                                    int*          exit_status)
{
	int tmp, pid;
	if (!exit_status)
		exit_status = &tmp;
	if (external->poll_fd.fd >= 0) {
		if (external->has_poll_fd)
			remove_poll_fd(external);
		close(external->poll_fd.fd);
		external->poll_fd.fd = -1;
	}
	if (external->reader_source) {
		guint source = external->reader_source;
		external->reader_source = 0;
		gdam_debug("external_kill_command: gsource=%d, external=%p,%d",
			source, external, GDAM_SOURCE(external)->source_id);
		g_source_remove(source);
	}
	if (external->process_id < 0)
		return FALSE;
	external->skipped = 0;
	pid = external->process_id;
	gdam_cleanup_unregister_child(pid);
	external->process_id = -1;
	if (waitpid(pid, exit_status, WNOHANG) > 0) {
		return TRUE;
	}
	kill(pid, SIGTERM);
	kill(pid, SIGKILL);
	while (waitpid(pid, exit_status, 0) < 0) {
		if (errno == EINTR) continue;
		perror("waitpid");
		return FALSE;
	}
	return TRUE;
}

static GSourceFuncs external_source_funcs = {
	gdam_external_prepare,
	gdam_external_check,
	gdam_external_dispatch,
	NULL
};

/*=====================================================================*
 *                      gdam_external_run_command                      *
 *                                                                     *
 * Start running a command in the gdam_external class.                 *
 *                                                                     *
 * TODO:  simulate `/bin/sh -c command' like perl does.                *
 *                                                                     *
 * Parameters:                                                         *
 *     external          the source which should run the command.      *
 *     sh_command        the command to run.                           *
 *=====================================================================*/
#define GDAM_EXTERNAL_PRIORITY       G_PRIORITY_DEFAULT
gboolean gdam_external_run_command(GdamExternal* external, 
                                   char**        command)
{
	int fds[2];

	if (external->process_id >= 0) {
		(void) gdam_external_kill_command(external, NULL);
	}

	g_return_val_if_fail(external->poll_fd.fd == -1, FALSE);
	g_return_val_if_fail(external->poll_fd.events ==
				(G_IO_IN|G_IO_ERR|G_IO_HUP), FALSE);
	g_return_val_if_fail(external->skipped == 0, FALSE);
	g_return_val_if_fail(external->process_id == -1, FALSE);
	g_return_val_if_fail(external->reader_source == 0, FALSE);

	if (pipe(fds) < 0) {
		perror("creating pipe");
		return FALSE;
	}
	gdam_buffer_construct(&external->buffer);
	if (external->command)
		g_strfreev(external->command);
	external->skipped = 0;
retry_fork:
	external->process_id = fork();
	if (external->process_id < 0) {
		if (errno == EAGAIN) { sleep(3); goto retry_fork; }
		if (errno == EINTR) goto retry_fork;
		perror("error forking");
		g_free(external);
		close(fds[0]);
		close(fds[1]);
		return FALSE;
	}
	if (external->process_id == 0) {
		GdamExternalClass* external_class;
		GtkObject* object;
		object = GTK_OBJECT(external);
		external_class = GDAM_EXTERNAL_CLASS(object->klass);

		dup2(fds[1], STDOUT_FILENO);
		(*external_class->do_exec)(external, command);

		_exit(127);
	}
	close(fds[1]);
	external->poll_fd.fd = fds[0];

	gdam_cleanup_register_child(external->process_id);
	external->reader_source = g_source_add(GDAM_EXTERNAL_PRIORITY,
	                                       TRUE,
		                               &external_source_funcs,
		                               NULL,
		                               external,
		                               NULL);
	return TRUE;
}
static gboolean do_one_read(GdamExternal* external)
{
	int n;
	char buf[32*1024];
redo_read:
	n = read(external->poll_fd.fd, buf, sizeof(buf));
	if (n < 0) {
		if (errno == EINTR)
			goto redo_read;
		if (errno == EWOULDBLOCK)
			return TRUE;
		perror("reading from external source");
		external->reader_source = 0;
		remove_poll_fd(external);
		return FALSE;
	}
	if (n == 0) {
		gdam_debug("end-of-file from external source "
					"[gsource %d, %p,%d]",
				external->reader_source, external,
				GDAM_SOURCE(external)->source_id);
		external->reader_source = 0;
		remove_poll_fd(external);
		return FALSE;
	}
	if (n <= external->skipped) {
		external->skipped -= n;
		return TRUE;
	} else if (external->skipped) {
		int s = external->skipped;
		gdam_buffer_append(&external->buffer, buf + s, n - s);
		external->skipped = 0;
		return TRUE;
	}

	gdam_buffer_append(&external->buffer, buf, n);
	return TRUE;
}

/*=====================================================================*
 *                          The GSourceFuncs.                          *
 *                                                                     *
 * These set up the PollFD (and indicate generally how                 *
 * to process the events involved.)                                    *
 *                                                                     *
 * AFAICT there is no convenient way to do this otherwise--            *
 * we'd have to add/remove watches all the time.                       *
 *=====================================================================*/

/* This function tells whether we know we need servicing 
 * before doing the poll.  It is also a good time to 
 * set up the PollFD.
 *
 * We always only need servicing if the fd is readable,
 * so we always return FALSE.
 *
 * But we do determine whether to poll at all via setting
 * external->poll_fd.events appropriately. */
static gboolean
gdam_external_prepare(gpointer  source_data,
                      GTimeVal* curtime,
                      gint*     timeout,
		      gpointer  user_data)
{
	GdamExternal* external = GDAM_EXTERNAL(user_data);

	g_return_val_if_fail(!GTK_OBJECT_DESTROYED(external), FALSE);

	/* We never have anything to do unless the source is ready. */
	*timeout = -1;

	/* Returning events == 0 means we have nothing to process. */
	external->poll_fd.events = external->poll_fd.revents = 0;
	
	/* We shouldn't be added as a source unless we have a valid
	 * file descriptor... */
	g_return_val_if_fail(external->poll_fd.fd >= 0, FALSE);

	/* Set up the read fd only if our buffer isn't
	 * too full already. */
	if (external->buffer.size >= external->max_buffer_size) {
		g_return_val_if_fail(external->skipped == 0, FALSE);
		external->poll_fd.events = 0;
		return FALSE;
	}

	if (!external->has_poll_fd)
		add_poll_fd(external);

	external->poll_fd.events = G_IO_IN|G_IO_ERR|G_IO_HUP;
	return FALSE;
}


/* Return whether we need servicing.
 * This is simple-- we just test revents. */
static gboolean
gdam_external_check(gpointer  source_data,
                    GTimeVal* curtime,
		    gpointer  user_data)
{
	GdamExternal* external = GDAM_EXTERNAL(user_data);
	g_return_val_if_fail(external->poll_fd.fd >= 0, FALSE);
	if (GDAM_SERVER_DEBUG_POLL)
		gdam_debug("gdam_external_check: revents=%d",
					external->poll_fd.revents);
	return external->poll_fd.revents ? TRUE : FALSE;
}
static gboolean
gdam_external_dispatch(gpointer  source_data,
                       GTimeVal* curtime,
		       gpointer  user_data)
{
	GdamExternal* external = GDAM_EXTERNAL(user_data);
	return do_one_read(external);
}


static gboolean
gdam_external_is_ready_to_read(GdamSource* source)
{
	GdamExternal* external = GDAM_EXTERNAL(source);
	int min_size = 1;

	/* Sources with no readers are either buffered,
	 * or at eof.  Never blocking on anything at least... */
	if (external->reader_source == 0)
		return TRUE;

	/* TEST: don't say we're ready to read unless we have 
	 * a 1/FRACTION_TO_WAIT_FOR our buffer
	 * or MAX_TO_WAIT_FOR samples, whichever is
	 * lower. */
	if (!external->has_been_ready) {
		min_size = external->max_buffer_size / FRACTION_TO_WAIT_FOR;
		min_size = MIN(min_size, MAX_TO_WAIT_FOR);
		if (external->buffer.size >= sizeof(GdamSample) * min_size) {
			external->has_been_ready = 1;
		}
	}

	return (external->buffer.size >= sizeof(GdamSample) * min_size);
}

static int
gdam_external_read(GdamSource* source, GdamSample* buf, int n)
{
	GdamExternal* external = GDAM_EXTERNAL(source);
	int rv = 0;
	if (external->zero_time > 0) {
		int i = MIN(n, external->zero_time);
		external->zero_time -= i;
		n -= i;
		gdam_sample_zero(buf, i);
		rv = i;
		buf += i;
	}

	if (external->buffer.size < sizeof(GdamSample)) {
		if (external->reader_source == 0)
			return rv;

		if (rv > 0)
			return rv;

		bzero(buf, sizeof(GdamSample));
		gdam_external_miss_data(source, 1);
		return rv + 1;
	}
	if (external->buffer.size / sizeof(GdamSample) < n)
		n = external->buffer.size / sizeof(GdamSample);
	if (external->samples_remaining >= 0) {
		if (n > external->samples_remaining)
			n = (int)external->samples_remaining;
		external->samples_remaining -= n;
	}
	gdam_buffer_read(&external->buffer,
	                 (char*)buf,
			 n * sizeof(GdamSample));
	external->samples_output += n;

	return rv + n;
}

static void gdam_external_miss_data(GdamSource* source, int n)
{
	int discarded;
	GdamExternal* external = GDAM_EXTERNAL(source);

	if (external->samples_remaining > 0) {
		if (n > external->samples_remaining)
			external->samples_remaining = 0;
		else
			external->samples_remaining -= n;
	}
	if (external->zero_time > 0) {
		int i = MIN(n, external->zero_time);
		external->zero_time -= i;
		n -= i;
	}

	external->samples_output += n;

	n *= sizeof(GdamSample);
	discarded = gdam_buffer_discard(&external->buffer, n);
	n -= discarded;		/* remaining bytes we cannot yet discard... */
	external->skipped += n;
}

static void gdam_external_destroy(GdamExternal* external)
{
	g_return_if_fail(external != NULL);
	gdam_external_kill_command(external, NULL);

	if (external->command) {
		g_strfreev(external->command);
		external->command = NULL;
	}

	gdam_buffer_destruct(&external->buffer);
	parent_class->destroy((GtkObject*)external);
}

static GdamTime gdam_external_get_position(GdamSource* source)
{
	GdamExternal* external = GDAM_EXTERNAL(source);
	return (gint64) external->samples_output - (gint64) external->zero_time;
}
