/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdammain.h"
#include "gdamwavetable.h"
#include "gdamnamespace.h"
#include <gtk/gtkmain.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>
#include "gdamconfigfile.h"
#include <gdam/gdambuffer.h>
#include <sys/stat.h>
#include <stdio.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>

char** gdam_config_filenames = NULL;

static void append_config_filename(const char* filename)
{
	int num_config_files = 0;
	char** at = gdam_config_filenames;
	if (!at) {
		gdam_config_filenames = g_new(char*, 2);
		gdam_config_filenames[0] = g_strdup(filename);
		gdam_config_filenames[1] = NULL;
		return;
	}
	while (*at) {
		num_config_files++;
		at++;
	}
	gdam_config_filenames = g_renew(char*,
	                                gdam_config_filenames,
					num_config_files + 2);
	gdam_config_filenames[num_config_files++] = g_strdup(filename);
	gdam_config_filenames[num_config_files] = NULL;
}

static const char*  config_file_locations[] = {
	GDAM_CONFIG_LOCATION,
	"server.config",
	NULL
};

static gboolean file_exists(const char* filename)
{
	struct stat sbuf;
	if (stat(filename, &sbuf) < 0)
		return FALSE;
	return TRUE;
}

static void find_default_config_file()
{
	int i;
	for (i = 0; config_file_locations[i] != NULL; i++) {
		const char* config = config_file_locations[i];
		if (file_exists(config)) {
			if (!gdam_config_file_parse(config)) {
				g_warning("error parsing config file %s",
				          config);
				return;
			}
			append_config_filename(config);
			return;
		}
	}
	g_error("no server config file was found");
}

static void daemon_log_func(const gchar*     log_domain,
                            GLogLevelFlags   log_level,
			    const gchar*     message,
			    gpointer         user_data)
{
	FILE* fp = gdam_log_fp;
	if (fp == NULL)
		fp = stderr;
	if ((log_level & G_LOG_LEVEL_MASK) == G_LOG_LEVEL_ERROR) {
		fprintf(fp, "*** ERROR ***: %s: %s.\n", log_domain, message);
	} else if ((log_level & G_LOG_LEVEL_MASK) == G_LOG_LEVEL_CRITICAL) {
		fprintf(fp, "*** %s-CRITICAL ***: %s.\n", log_domain, message);
	} else if ((log_level & G_LOG_LEVEL_MASK) == G_LOG_LEVEL_WARNING) {
		fprintf(fp, "*** WARNING ***: %s: %s.\n", log_domain, message);
	} else if ((log_level & G_LOG_LEVEL_MASK) == G_LOG_LEVEL_MESSAGE) {
		fprintf(fp, "Message: %s: %s.\n", log_domain, message);
	} else if ((log_level & G_LOG_LEVEL_MASK) == G_LOG_LEVEL_INFO) {
		fprintf(fp, "Info: %s: %s.\n", log_domain, message);
	} else if ((log_level & G_LOG_LEVEL_MASK) == G_LOG_LEVEL_DEBUG) {
		fprintf(fp, "debug: %s: %s.\n", log_domain, message);
	}
}
static void maybe_write_pid_file(int pid)
{
}

static void daemonize()
{
	int uid, gid;
	int pid;
	struct passwd* user_info;
	const char* username;

	chdir("/");
	umask(0);

	/* Redirect glib's logging facility. */
	g_log_set_handler("gdam", G_LOG_LEVEL_INFO
		                 |G_LOG_LEVEL_DEBUG
		                 |G_LOG_LEVEL_ERROR
		                 |G_LOG_LEVEL_WARNING
		                 |G_LOG_LEVEL_MESSAGE,
			   daemon_log_func,
			   NULL);

	if (getuid() == 0) {
		username = gdam_username ? gdam_username : "nobody";
		user_info = getpwnam(username);
		if (user_info == NULL)
			g_error("couldn't get user-id for %s", username);
		uid = user_info->pw_uid;
		gid = user_info->pw_gid;
		if (gdam_groupname != NULL) {
			struct group* group_info;
			group_info = getgrnam(gdam_groupname);
			if (group_info == NULL)
				g_error("couldn't get group-id for %s",
					gdam_groupname);
			gid = group_info->gr_gid;
		}
#if HAVE_SETREGID
		if (setregid(gid, gid) < 0)
			g_error("setregid(%d): %s", gid, strerror(errno));
#else
		if (setgid(gid) < 0)
			g_error("setgid(%d): %s", gid, strerror(errno));
#endif
#if HAVE_SETREUID
		if (setreuid(uid, uid) < 0)
			g_error("setreuid(%d): %s", uid, strerror(errno));
#else
		if (setuid(uid) < 0)
			g_error("setuid(%d): %s", uid, strerror(errno));
#endif
	} else {
		if (gdam_username)
			g_warning("you aren't root, so `User %s' ignored",
				gdam_username);
		if (gdam_groupname)
			g_warning("you aren't root, so `Group %s' ignored",
				gdam_groupname);
	}

retry_fork:
	pid = fork();
	if (pid < 0) {
		if (errno == EINTR)
			goto retry_fork;
		if (errno == EAGAIN) {
			sleep(5);
			goto retry_fork;
		}
		g_error("error forking: %s", strerror(errno));
	}
	if (pid > 0) {
		maybe_write_pid_file(pid);
		_exit(0);
	}

}

void gdam_init(int* argc,
               char*** argv)
{
	/* Probably shouldn't really bother doing this... */
	extern void gtk_object_post_arg_parsing_init();
	int i, out_at;

	gtk_type_init();

	gtk_object_post_arg_parsing_init();	/* nor this... */

	gtk_signal_init();

	for (i = 1; i < *argc; i++) {
		char* arg = (*argv)[i];
		char* param = (*argv)[i+1];

		if (strcmp(arg, "--debug") == 0) {
			gdam_debugging_on = TRUE;
			gdam_debug("debugging on");
			(*argv)[i] = 0;
			continue;
		}

		if (strcmp(arg, "--config") == 0) {
			if (!param) {
				g_error("--config specified without "
				        "a config filename");
			}
			if (!gdam_config_file_parse(param)) {
				g_warning("error parsing config file %s",
				          param);
				return;
			}
			gdam_debug("Parsed config file %s", param);
			(*argv)[i] = (*argv)[i+1] = NULL;
			i++;
			append_config_filename(param);
			continue;
		}
	}

	if (gdam_config_filenames == NULL) {
		/* Search default configfile list. */
		find_default_config_file();
	}

	if (gdam_daemonize)
		daemonize();

	gdam_wavetable_init_common();
#if 0
	/* Add a few common server-side ints. */
	gdam_namespace_int_set("wavetables",
	                       "sine",
			       gdam_wavetable_sine_id());
	gdam_namespace_int_set("wavetables",
	                       "triangle",
			       gdam_wavetable_triangle_id());
	gdam_namespace_int_set("wavetables",
	                       "square",
			       gdam_wavetable_square_id());
#endif

	/* Condense deleted arguments. */
	for (i = out_at = 1; i < *argc; i++)
		if ((*argv)[i] != NULL) {
			if (out_at == i)
				out_at++;
			else
				(*argv)[out_at++] = (*argv)[i];
		}
	
	*argc = out_at;
}


void gdam_main()
{
	gtk_main();
}

