/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#ifndef __GDAM_SOUNDSOURCE_MIX_H_
#define __GDAM_SOUNDSOURCE_MIX_H_

#include <glib.h>
#include "gdamsource.h"
#include <gdam/gdambuffer.h>	/* for GdamTime */

GtkType gdam_mix_get_type();
#define GDAM_TYPE_MIX              (gdam_mix_get_type ())
#define GDAM_MIX(obj)              (GTK_CHECK_CAST ((obj), GDAM_TYPE_MIX, GdamMix))
#define GDAM_MIX_CLASS(klass)      (GTK_CHECK_CLASS_CAST ((klass), GDAM_TYPE_MIX, GdamSoundSourceClass))
#define GDAM_IS_MIX(obj)           (GTK_CHECK_TYPE ((obj), GDAM_TYPE_MIX))
#define GDAM_IS_MIX_CLASS(klass)   (GTK_CHECK_CLASS_TYPE ((klass), GDAM_TYPE_MIX))

typedef struct _GdamMix GdamMix;
typedef struct _GdamMixClass GdamMixClass;
typedef struct _GdamMixElement GdamMixElement;
typedef enum _GdamMixMode GdamMixMode;

enum _GdamMixMode {
	GDAM_MIX_ETERNAL,
	GDAM_MIX_FIRST_ENDED,
	GDAM_MIX_ALL_ENDED
};

struct _GdamMixElement {
	/* The next sample (in the mix's perception of time)
	 * to fill from the subsource. */
	GdamTime		read_to;

	/* The source to read from. */
	GdamSource*		sub_source;

	/* We use an inplace list, b/c we want deleting
	 * a GdamMixElement to be fast... */
	GdamMixElement*		next;
	GdamMixElement*		prev;

	/* Passed to the init_copy and augment functions. */
	gpointer                source_info;

	/* Optional name. */
	char*			name;

	gboolean		misses_data;

	unsigned		flags;
};

/* For GdamMixElement::flags. */
#define MIX_ELEMENT_IN_FUTURE		1

struct _GdamMix {
	GdamSource		base;

	GdamMixElement*		first_source;
	GdamMixElement*		last_source;

	/* A circular buffer we are filling up. */
	int			buffer_size;
	GdamSample*		buffer;
	int			buffer_data_start;

	/* A temporary buffer, the same length as `buffer' (ie buffer_size)
	 * that is used only during `read'.  Just saves
	 * us a `malloc' during read... */
	GdamSample*		tmp_buf;

	/* Keeps track of how many samples in the buffer have been
	 * initialized. */
	GdamTime 		cleared_to;

	GdamTime		samples_output;

	/* Time to stop outputting at. (used internally to implement
	 * GDAM_MIX_FIRST_ENDED) */
	GdamTime		samples_remaining;

	/* Last time we filled (used for GDAM_MIX_ALL_ENDED). */
	GdamTime		last_read_to;

	/* Mapping from names to named elements. */
	GHashTable*		names;

	GdamMixMode		mode;
};

struct _GdamMixClass {
	GdamSourceClass klass;
	void        (*connect)(GdamMix*, GdamSource* to_add, gint64 when);

	/* A function which controls how the mix is modified when
	 * adding/starting. */
	void        (*initial_copy)(GdamSample*          destination,
	                            const GdamSample*    source,
				    int                  num_samples,
				    gpointer             source_info);
	void        (*augment)     (GdamSample*          destination,
	                            const GdamSample*    source,
				    int                  num_samples,
				    gpointer             source_info);
};


GdamMix*       gdam_mix_new(gboolean terminates_when_done);

/* If `when' is `-1' it means `immediately'. */
void           gdam_mix_connect         (GdamMix*    mix,
                                         GdamSource* sub_source,
				         GdamTime    when);

/* When is ignored if the source already exists. */
void           gdam_mix_connect_named   (GdamMix*    mix,
                                         GdamSource* sub_source,
				         GdamTime    when,
					 const char* name);
GdamSource*    gdam_mix_get_named       (GdamMix*    mix,
                                         const char* name);

GdamTime       gdam_mix_get_pos         (GdamMix*    mix);
GdamTime       gdam_mix_get_sub_pos     (GdamMix*    mix,
                                         GdamSource* sub_source);
int            gdam_mix_num_sources     (GdamMix*    mix);

/* These functions provided so modulator can use them... */
void gdam_mix_arg_set_mixed(GtkObject* object, GdamArg* arg);
void gdam_mix_arg_get_mixed(GtkObject* object, GdamArg* arg);

#endif
