/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamproxysource.h"
#include <gdam/gdam-protocol.h>
#include "gdamconnection.h"

static void gdam_proxy_source_class_init       (GdamSourceClass*  klass);
static void gdam_proxy_source_init             (GdamProxySource*  proxy);
static void gdam_proxy_source_destroy          (GtkObject*        proxy);
static GdamTime 
            gdam_proxy_source_get_position     (GdamSource*       input);
static int  gdam_proxy_source_read             (GdamSource*       input,
                                                GdamSample*       samples,
		                                int               max_read);
static void gdam_proxy_source_miss_data        (GdamSource*       input,
                                                int               samples_lost);
static gboolean
            gdam_proxy_source_is_ready_to_read (GdamSource* proxy_source);

static GtkObjectClass* parent_class = NULL;

GtkType gdam_proxy_source_get_type()
{
	static GtkType proxy_source_type = 0;
	if (!proxy_source_type) {
		static const GtkTypeInfo proxy_source_info = {
			"GdamProxySource",
			sizeof(GdamProxySource),
			sizeof(GdamSourceClass),
			(GtkClassInitFunc) gdam_proxy_source_class_init,
			(GtkObjectInitFunc) gdam_proxy_source_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		parent_class = gtk_type_class(GDAM_TYPE_SOURCE);
		proxy_source_type = gtk_type_unique(GDAM_TYPE_SOURCE,
		                                    &proxy_source_info);
	}
	return proxy_source_type;
}

static void
gdam_proxy_source_get_fade_envelope(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = GDAM_PROXY_SOURCE(object)->max_envelope;
}
static void
gdam_proxy_source_set_fade_envelope(GtkObject* object, GdamArg* arg)
{
	gdam_proxy_source_set_envelope(GDAM_PROXY_SOURCE(object),
	                               GDAM_ARG_INT(arg));
}

static void gdam_proxy_source_class_init(GdamSourceClass* klass)
{
	GtkObjectClass* object_class = (GtkObjectClass*)klass;
	object_class->destroy = gdam_proxy_source_destroy;
	klass->read = gdam_proxy_source_read;
	klass->miss_data = gdam_proxy_source_miss_data;
	klass->is_ready_to_read = gdam_proxy_source_is_ready_to_read;
	klass->get_position = gdam_proxy_source_get_position;
	gdam_arg_register(object_class,
	                  "fade_envelope",
			  GDAM_ARG_TYPE_INT,
			  gdam_proxy_source_set_fade_envelope,
			  gdam_proxy_source_get_fade_envelope,
			  GDAM_ARG_VITAL);
}

static void gdam_proxy_source_init(GdamProxySource* proxy_source)
{
}

static void gdam_proxy_source_destroy(GtkObject* proxy)
{
	GdamProxySource* proxy_source = GDAM_PROXY_SOURCE(proxy);
	if (proxy_source->delegate) {
		GtkObject* delegate = GTK_OBJECT(proxy_source->delegate);
		proxy_source->delegate = NULL;
		gtk_object_unref(delegate);
	}
	if (proxy_source->cueing) {
		GtkObject* cueing = GTK_OBJECT(proxy_source->cueing);
		proxy_source->cueing = NULL;
		gtk_object_unref(cueing);
	}
	(*parent_class->destroy)(proxy);
}

static int gdam_proxy_source_read(GdamSource* input,
                                  GdamSample* samples,
		                  int max_read)
{
	GdamProxySource* proxy_source = GDAM_PROXY_SOURCE(input);
	GdamSource* delegate = proxy_source->delegate;
	int rv;
	int fade_index = proxy_source->fading_out_index;
	int max_fade_index = proxy_source->max_envelope;
	if (!delegate)
		return 0;
	rv = gdam_source_read(delegate, samples, max_read);
	if (rv > 0 && proxy_source->cueing)
		gdam_source_miss_data(proxy_source->cueing, rv);
	if (fade_index < max_fade_index) {
		///int i = 0;
		GdamFixed start_vol, delta_vol;
		GdamSample* src = proxy_source->fading_out + fade_index;
		GdamSample* dst = samples;
		start_vol = fade_index * (1<<14) / max_fade_index;
		delta_vol = (1<<14) / max_fade_index;
		{
			int to_fade = (max_fade_index - fade_index);
			if (to_fade > rv)
				to_fade = rv;
			while (to_fade > 0) {
				int tmp;
				tmp = (    (start_vol          * dst->s[0])
					+ (((1<<14)-start_vol) * src->s[0])  );
				tmp >>= 14;
				dst->s[0] = CLAMP(tmp,
				                  GDAM_SAMPLE_MIN,
						  GDAM_SAMPLE_MAX);
				tmp = (    (start_vol          * dst->s[1])
					+ (((1<<14)-start_vol) * src->s[1])  );
				tmp >>= 14;
				dst->s[1] = CLAMP(tmp,
				                  GDAM_SAMPLE_MIN,
						  GDAM_SAMPLE_MAX);
				dst++;
				src++;
				to_fade--;
				fade_index++;
				start_vol += delta_vol;
			}
		}
		proxy_source->fading_out_index = fade_index;
	}
	return rv;
}
static GdamTime gdam_proxy_source_get_position(GdamSource* input)
{
	GdamProxySource* proxy_source = GDAM_PROXY_SOURCE(input);
	GdamSource* delegate = proxy_source->delegate;
	if (!delegate)
		return -1;
	if (proxy_source->cueing)
		return gdam_source_get_position(proxy_source->cueing);
	return gdam_source_get_position(delegate);
}

static void gdam_proxy_source_miss_data(GdamSource* input,
                                        int samples_lost)
{
	GdamProxySource* proxy_source = GDAM_PROXY_SOURCE(input);
	GdamSource* delegate = proxy_source->delegate;
	if (delegate)
		gdam_source_miss_data(delegate, samples_lost);
	if (proxy_source->cueing)
		gdam_source_miss_data(proxy_source->cueing, samples_lost);
}

static gboolean gdam_proxy_source_is_ready_to_read(GdamSource* proxy)
{
	GdamProxySource* proxy_source = GDAM_PROXY_SOURCE(proxy);
	GdamSource* delegate = proxy_source->delegate;
	GdamSource* cueing = proxy_source->cueing;
	if (cueing && gdam_source_is_ready_to_read(cueing)) {
		if (delegate) {
			proxy_source->delegate = NULL;
			if (proxy_source->max_envelope > 0) {
				int num_read;
				int slide;
				num_read = gdam_source_read
				                 (delegate,
				                  proxy_source->fading_out,
						  proxy_source->max_envelope);
				if (num_read < 0) num_read = 0;
				slide = proxy_source->max_envelope - num_read;
				if (slide > 0)
					memmove(proxy_source->fading_out+slide,
					        proxy_source->fading_out,
						sizeof(GdamSample) * num_read);
				proxy_source->fading_out_index = slide;
			}
			gtk_object_unref(GTK_OBJECT(delegate));
		}
		gdam_debug("cued ss becoming delegate");
		proxy_source->delegate = cueing;
		proxy_source->cueing = NULL;
		return TRUE;
	}
	if (!delegate)
		return TRUE;
	return gdam_source_is_ready_to_read(delegate);
}


void   gdam_proxy_source_set    (GdamProxySource* proxy_source,
                                 GdamSource*      delegate)
{
	g_return_if_fail(proxy_source != NULL);
	g_return_if_fail(delegate != NULL);
	gtk_object_ref(GTK_OBJECT(delegate));
	if (proxy_source->delegate) {
		GtkObject* old_delegate = GTK_OBJECT(proxy_source->delegate);
		proxy_source->delegate = NULL;
		gtk_object_unref(old_delegate);
	}
	gtk_object_sink(GTK_OBJECT(delegate));
	proxy_source->delegate = delegate;
}

void   gdam_proxy_source_set_cue(GdamProxySource* proxy_source,
                                 GdamSource*      to_cue)
{
	g_return_if_fail(proxy_source != NULL);
	g_return_if_fail(to_cue != NULL);
	gtk_object_ref(GTK_OBJECT(to_cue));
	if (proxy_source->cueing) {
		GtkObject* old_cueing = GTK_OBJECT(proxy_source->cueing);
		proxy_source->cueing = NULL;
		gtk_object_unref(old_cueing);
	}
	gtk_object_sink(GTK_OBJECT(to_cue));
	proxy_source->cueing = to_cue;
}
void   gdam_proxy_source_set_envelope(GdamProxySource* proxy_source,
                                      int              max_envelope)
{
	proxy_source->max_envelope = max_envelope;
	proxy_source->fading_out = g_renew(GdamSample,
	                                   proxy_source->fading_out,
					   max_envelope);
	proxy_source->fading_out_index = max_envelope;

}
