/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamserver.h"
#include "gdamconnection.h"
#include "gdamserverdebug.h"
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <errno.h>

#include <glib.h>

static gboolean    gdam_server_prepare          (gpointer         src_data,
                                                 GTimeVal*        curtime,
                                                 gint*            timeout,
		                                 gpointer         user_data);
static gboolean    gdam_server_check            (gpointer         src_data,
                                                 GTimeVal*        curtime,
		                                 gpointer         user_data);
static gboolean    gdam_server_dispatch         (gpointer         src_data,
                                                 GTimeVal*        curtime,
		                                 gpointer         user_data);

static GSourceFuncs gdam_server_source_funcs = {
	gdam_server_prepare,
	gdam_server_check,
	gdam_server_dispatch,
	NULL
};

static gboolean gdam_server_prepare(gpointer         src_data,
                                    GTimeVal*        curtime,
                                    gint*            timeout,
		                    gpointer         user_data)
{
	return FALSE;
}
static gboolean gdam_server_check(gpointer         src_data,
                                  GTimeVal*        curtime,
		                  gpointer         user_data)
{
	GdamServer* server = (GdamServer*)user_data;
	if (GDAM_SERVER_DEBUG_POLL)
		gdam_debug("gdam_server_check: revents=%d", server->poll_fd.revents);
	return server->poll_fd.revents ? TRUE : FALSE;
}

/* Return TRUE if you want to remain a source. */
static gboolean gdam_server_dispatch(gpointer         src_data,
                                     GTimeVal*        curtime,
		                     gpointer         user_data)
{
	GdamServer* server = (GdamServer*)user_data;
	GdamConnection* accepted_connection;

	/* Accept a new connection (at the socket layer.) */
	int listening_fd = server->poll_fd.fd;
	int accepted_fd;
	struct sockaddr_in sin;
	int sin_size = sizeof(struct sockaddr_in);
	if (listening_fd <= 0)
		return FALSE;
	bzero(&sin, sizeof(sin));
	accepted_fd = accept(listening_fd, (struct sockaddr*)&sin, &sin_size);
	if (accepted_fd < 0) {
		if (errno == EWOULDBLOCK || errno == EINTR)
			return TRUE;
		g_warning("gdam_server_dispatch: "
			  "error accepting new connection: %s",
			  strerror(errno));
		server->source_id = 0;
		return FALSE;
	}

	/* Make a connection and attach it to the main loop. */
	accepted_connection = gdam_connection_new(accepted_fd,
	                                          server->authenticator);
	if (!accepted_connection) {
		g_warning("failed to make new connection");
		server->source_id = 0;
		return FALSE;
	}

	gdam_debug("connection %p on fd %d attached to mainloop",
		   accepted_connection, accepted_fd);

	gdam_connection_attach_to_main_loop(accepted_connection);

	return TRUE;
}

GdamServer* gdam_server_new(int sock_fd, GdamAuthenticator* auth)
{
	GdamServer* server = g_new(GdamServer, 1);
	server->authenticator = auth;
	server->poll_fd.fd = sock_fd;
	server->poll_fd.events = G_IO_IN;
	server->source_id = 0;
	server->is_destructing = 0;
	return server;
}

void gdam_server_attach(GdamServer* server)
{
	g_return_if_fail(server->source_id == 0);
	g_main_add_poll(&server->poll_fd, G_PRIORITY_LOW);

	if (server->source_id != 0)
		return;

	server->source_id = g_source_add(G_PRIORITY_LOW, FALSE,
	                                 &gdam_server_source_funcs,
					 NULL,
					 server,
					 (GDestroyNotify) gdam_server_destroy);
}

void gdam_server_detach(GdamServer* server)
{
	if (server->source_id) {
		int source_id = server->source_id;
		server->source_id = 0;
		g_source_remove(source_id);
	}
	g_main_remove_poll(&server->poll_fd);
}

void gdam_server_destroy(GdamServer* server)
{
	if (server->is_destructing)
		return;
	server->is_destructing = 1;
	gdam_server_detach(server);
	if (server->poll_fd.fd >= 0) {
		close(server->poll_fd.fd);
		server->poll_fd.fd = -1;
	}
	g_free(server);
}

