/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamrecorder.h"
#include <gdam/gdamcachebuffer.h>

static GtkObjectClass* parent_class = NULL;

static int  gdam_recorder_read(GdamSource* source,
                               GdamSample* buffer,
			       int         max_read)
{
	GdamSource* filtered = GDAM_FILTER(source)->filtered;
	GdamRecorder* recorder = GDAM_RECORDER(source);
	int num_read;
	if (filtered == NULL)
		return 0;
	num_read = gdam_source_read(filtered, buffer, max_read);
	if (recorder->is_recording && num_read > 0)
		gdam_buffer_append(&recorder->buffer,
		                   buffer,
				   sizeof(GdamSample) * max_read);
	return num_read;
}
static void  gdam_recorder_destroy(GtkObject* recorder)
{
	GdamBuffer* buffer = &(GDAM_RECORDER(recorder)->buffer);
	gdam_buffer_destruct(buffer);
	(*parent_class->destroy)(recorder);
}

static void gdam_recorder_init(GdamRecorder* recorder)
{
	gdam_buffer_construct(&recorder->buffer);
}
static void gdam_recorder_arg_set_recording(GtkObject* object, GdamArg* arg)
{
	GDAM_RECORDER(object)->is_recording = GDAM_ARG_INT(arg);
}

static void gdam_recorder_arg_get_recording(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = GDAM_RECORDER(object)->is_recording;
}

static void gdam_recorder_arg_copy_buffer(GtkObject* object, GdamArg* arg)
{
	GdamRecorder* recorder = GDAM_RECORDER(object);
	GdamCacheBuffer* buffer = gdam_recorder_make_buffer(recorder, FALSE);
	if (gdam_connection_for_arg != NULL)
		gdam_connection_ref(gdam_connection_for_arg,
		                    GTK_OBJECT(buffer));
	GDAM_ARG_INT(arg) = buffer->buffer_id;
}

static void gdam_recorder_arg_grab_buffer(GtkObject* object, GdamArg* arg)
{
	GdamRecorder* recorder = GDAM_RECORDER(object);
	GdamCacheBuffer* buffer = gdam_recorder_make_buffer(recorder, TRUE);
	if (gdam_connection_for_arg != NULL)
		gdam_connection_ref(gdam_connection_for_arg,
		                    GTK_OBJECT(buffer));
	GDAM_ARG_INT(arg) = buffer->buffer_id;
}

static void gdam_recorder_arg_get_length(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = GDAM_RECORDER(object)->buffer.size;
}


static void gdam_recorder_class_init(GdamRecorderClass* recorder_class)
{
	GdamSourceClass* source_class = GDAM_SOURCE_CLASS(recorder_class);
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(recorder_class);
	source_class->read = gdam_recorder_read;
	object_class->destroy = gdam_recorder_destroy;

	gdam_arg_register(object_class,
	                  "recording",
			  GDAM_ARG_TYPE_INT,
			  gdam_recorder_arg_set_recording,
			  gdam_recorder_arg_get_recording,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "copy_buffer",
			  GDAM_ARG_TYPE_INT,
			  NULL,
			  gdam_recorder_arg_copy_buffer,
			  GDAM_ARG_TRIGGER);
	gdam_arg_register(object_class,
	                  "grab_buffer",
			  GDAM_ARG_TYPE_INT,
			  NULL,
			  gdam_recorder_arg_grab_buffer,
			  GDAM_ARG_TRIGGER);
	gdam_arg_register(object_class,
	                  "length",
			  GDAM_ARG_TYPE_INT,
			  NULL,
			  gdam_recorder_arg_get_length,
			  0);
}

GtkType gdam_recorder_get_type()
{
	static GtkType recorder_type = 0;
	if (!recorder_type) {
		static const GtkTypeInfo recorder_info = {
			"GdamRecorder",
			sizeof(GdamRecorder),
			sizeof(GdamRecorderClass),
			(GtkClassInitFunc) gdam_recorder_class_init,
			(GtkObjectInitFunc) gdam_recorder_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_FILTER;
		parent_class = gtk_type_class(parent);
		recorder_type = gtk_type_unique(parent,
		                                 &recorder_info);
	}
	return recorder_type;
}

void gdam_recorder_stop(GdamRecorder* recorder)
{
	recorder->is_recording = FALSE;
}

void gdam_recorder_start(GdamRecorder* recorder)
{
	recorder->is_recording = TRUE;
}

void gdam_recorder_empty(GdamRecorder* recorder)
{
	gdam_buffer_destruct(&recorder->buffer);
}

GdamCacheBuffer*
gdam_recorder_make_buffer(GdamRecorder* recorder, gboolean drain)
{
	GdamSample* buf;
	int num_samples = recorder->buffer.size / sizeof(GdamSample);
	GdamCacheBuffer* rv;
	gdam_debug("gdam_recorder_make_buffer: %d samples", num_samples);
	buf = g_new(GdamSample, num_samples);
	if (drain)
		gdam_buffer_read(&recorder->buffer,
		                 buf,
				 sizeof(GdamSample) * num_samples);
	else
		gdam_buffer_peek(&recorder->buffer,
		                 buf,
				 sizeof(GdamSample) * num_samples);
	rv = gdam_cache_buffer_memory(num_samples, buf, g_free);
	g_return_val_if_fail(rv != NULL, NULL);
	return rv;
}

