#include "gdamsecondorderfilter.h"
#include <math.h>

static GtkObjectClass* parent_class = NULL;

static void gdam_second_order_filter_init(GdamSecondOrderFilter* second_order_filter)
{
}

static GdamFixed* convert_to_fixed(GdamArg* arg)
{
	int length = GDAM_ARG_DOUBLE_ARRAY(arg).length;
	const gdouble* array = GDAM_ARG_DOUBLE_ARRAY(arg).array;
	GdamFixed* rv = g_new(GdamFixed, length);
	GdamFixed* at = rv;
	while (length-- > 0)
		*at++ = (GdamFixed) (*array++ * GdamFixed_ONE);
	return rv;
}

static gdouble* convert_to_doubles(const GdamFixed* array, int length)
{
	gdouble* rv = g_new(gdouble, length);
	gdouble* at = rv;
	while (length-- > 0)
		*at++ = (gdouble) (*array++) / GdamFixed_ONE;
	return rv;
}

static void compute_max_length(GdamSecondOrderFilter* filter)
{
	int old_max_length = filter->max_length;
	int max_length = filter->in_first_order_length;
	if (max_length < filter->in_second_order_length)
		max_length = filter->in_second_order_length;
	if (max_length < filter->out_first_order_length)
		max_length = filter->out_first_order_length;
	if (max_length < filter->out_second_order_length)
		max_length = filter->out_second_order_length;
	if (max_length != old_max_length) {
		filter->past_in = g_renew(GdamSample,
		                          filter->past_in,
					  max_length);
		filter->past_out = g_renew(GdamSample,
		                           filter->past_in,
					   max_length);
		if (max_length > old_max_length) {
			gdam_sample_zero(filter->past_in + old_max_length,
			                 max_length - old_max_length);
			gdam_sample_zero(filter->past_out + old_max_length,
			                 max_length - old_max_length);
		}
	}
	filter->max_length = max_length;
}

static void gdam_second_arg_set_first_order(GtkObject* object, GdamArg* arg)
{
	GdamSecondOrderFilter* filter;
	gboolean is_in;

	filter = GDAM_SECOND_ORDER_FILTER(object);

	if (strcmp(arg->name, "first_order_in") == 0)
		is_in = TRUE;
	else
		is_in = FALSE;

	if (is_in) {
		g_free(filter->in_first_order);
		filter->in_first_order = convert_to_fixed(arg);
		filter->in_first_order_length
					= GDAM_ARG_DOUBLE_ARRAY(arg).length;
	} else {
		g_free(filter->out_first_order);
		filter->out_first_order = convert_to_fixed(arg);
		filter->out_first_order_length
					= GDAM_ARG_DOUBLE_ARRAY(arg).length;
	}
	compute_max_length(filter);
}

static void gdam_second_arg_get_first_order(GtkObject* object, GdamArg* arg)
{
	GdamSecondOrderFilter* filter;
	int length;
	gboolean is_in;
	GdamFixed* array;

	filter = GDAM_SECOND_ORDER_FILTER(object);

	if (strcmp(arg->name, "first_order_in") == 0)
		is_in = TRUE;
	else
		is_in = FALSE;

	if (is_in) {
		array = filter->in_first_order;
		length = filter->in_first_order_length;
	} else {
		array = filter->out_first_order;
		length = filter->out_first_order_length;
	}
	
	GDAM_ARG_DOUBLE_ARRAY(arg).length = length;
	GDAM_ARG_DOUBLE_ARRAY(arg).array = convert_to_doubles(array, length);
}

static void gdam_second_arg_set_second_order(GtkObject* object, GdamArg* arg)
{
	GdamSecondOrderFilter* filter;
	gboolean is_in;
	int array_length;
	int length;

	filter = GDAM_SECOND_ORDER_FILTER(object);

	array_length = GDAM_ARG_DOUBLE_ARRAY(arg).length;
	length = (int) sqrt(array_length);

	if (array_length != length * length) {
		gdam_debug("%s must be a square array", arg->name);
		return;
	}

	if (strcmp(arg->name, "second_order_in") == 0)
		is_in = TRUE;
	else
		is_in = FALSE;

	if (is_in) {
		g_free(filter->in_second_order);
		filter->in_second_order = convert_to_fixed(arg);
		filter->in_second_order_length = length;
	} else {
		g_free(filter->out_second_order);
		filter->out_second_order = convert_to_fixed(arg);
		filter->out_second_order_length = length;
	}
}

static void gdam_second_arg_get_second_order(GtkObject* object, GdamArg* arg)
{
	GdamSecondOrderFilter* filter;
	gboolean is_in;
	int length;
	int array_length;
	GdamFixed* array;

	filter = GDAM_SECOND_ORDER_FILTER(object);

	if (strcmp(arg->name, "second_order_in") == 0)
		is_in = TRUE;
	else
		is_in = FALSE;

	if (is_in) {
		array = filter->in_second_order;
		length = filter->in_second_order_length;
	} else {
		array = filter->out_second_order;
		length = filter->out_second_order_length;
	}
	
	GDAM_ARG_DOUBLE_ARRAY(arg).length = length;
	array_length = length * length;
	GDAM_ARG_DOUBLE_ARRAY(arg).array
			= convert_to_doubles(array, array_length);
}


static gint16
process_channel(GdamSecondOrderFilter*       filter,
                int                          channel)
{
	GdamSample* past_in = filter->past_in;
	GdamSample* past_out = filter->past_out;
	GdamFixed* second_order_filter_at;
	int i;

	gint64 accum = 0;

	for (i = 0; i < filter->in_first_order_length; i++)
		accum += (gint64) past_in[i].s[channel]
		                * filter->in_first_order[i];

	for (i = 0; i < filter->out_first_order_length; i++)
		accum += (gint64) past_out[i].s[channel]
		                * filter->out_first_order[i];

	accum *= GDAM_SAMPLE_MAX;

	second_order_filter_at = filter->in_second_order;

	for (i = 0; i < filter->in_second_order_length; i++) {
		int j;
		second_order_filter_at += i;
		for (j = 0; j < filter->in_second_order_length; j++)
			accum += (gint64) past_in[i].s[channel]
			       * (gint64) past_in[j].s[channel]
			       * (*second_order_filter_at++);
	}

	second_order_filter_at = filter->out_second_order;

	for (i = 0; i < filter->out_second_order_length; i++) {
		int j;
		second_order_filter_at += i;
		for (j = 0; j < filter->out_second_order_length; j++)
			accum += (gint64) past_out[i].s[channel]
			       * (gint64) past_out[j].s[channel]
			       * (*second_order_filter_at++);
	}

	accum /= (gint64)GDAM_SAMPLE_MAX * GDAM_SAMPLE_MAX;
	return CLAMP((int)accum, GDAM_SAMPLE_MIN, GDAM_SAMPLE_MAX);
}

static inline GdamSample
process_sample(GdamSecondOrderFilter*       filter,
               GdamSample                   sample)
{
	GdamSample out;

	if (filter->max_length >= 1) {
		memmove(filter->past_in + 1,
			filter->past_in,
			(filter->max_length - 1) * sizeof(GdamSample));
		filter->past_in[0] = sample;
	}

	out.s[0] = process_channel(filter, 0);
	out.s[1] = process_channel(filter, 1);

	if (filter->max_length >= 1) {
		memmove(filter->past_out + 1,
			filter->past_out,
			(filter->max_length - 1) * sizeof(GdamSample));
		filter->past_out[0] = out;
	}

	return out;
}

static int gdam_second_order_filter_read(GdamSource* source,
                                         GdamSample* buffer,
					 int         num_samples)
{
	int num_read;
	int i;
	GdamSource* filtered;
	GdamSecondOrderFilter* filter;

	filtered = GDAM_FILTER(source)->filtered;
	filter = GDAM_SECOND_ORDER_FILTER(source);

	num_read = gdam_source_read(filtered, buffer, num_samples);

	for (i = 0; i < num_read; i++)
		buffer[i] = process_sample(filter, buffer[i]);

	return num_read;
}

static void gdam_second_order_filter_destroy(GtkObject* object)
{
	GdamSecondOrderFilter* filter = GDAM_SECOND_ORDER_FILTER(object);
	g_free(filter->in_first_order);
	g_free(filter->out_first_order);
	g_free(filter->in_second_order);
	g_free(filter->out_second_order);
	g_free(filter->past_in);
	g_free(filter->past_out);
	(*parent_class->destroy)(object);
}

static void gdam_second_order_filter_class_init(GdamSourceClass* source_class)
{
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(source_class);
	object_class->destroy = gdam_second_order_filter_destroy;
	source_class->read = gdam_second_order_filter_read;

	gdam_arg_register(object_class,
	                  "second_order_in",
			  GDAM_ARG_TYPE_DOUBLE_ARRAY,
			  gdam_second_arg_set_second_order,
			  gdam_second_arg_get_second_order,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "second_order_out",
			  GDAM_ARG_TYPE_DOUBLE_ARRAY,
			  gdam_second_arg_set_second_order,
			  gdam_second_arg_get_second_order,
			  GDAM_ARG_VITAL);

	gdam_arg_register(object_class,
	                  "first_order_in",
			  GDAM_ARG_TYPE_DOUBLE_ARRAY,
			  gdam_second_arg_set_first_order,
			  gdam_second_arg_get_first_order,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "first_order_out",
			  GDAM_ARG_TYPE_DOUBLE_ARRAY,
			  gdam_second_arg_set_first_order,
			  gdam_second_arg_get_first_order,
			  GDAM_ARG_VITAL);
}

GtkType gdam_second_order_filter_get_type()
{
	static GtkType second_order_filter_type = 0;
	if (!second_order_filter_type) {
		static const GtkTypeInfo second_order_filter_info = {
			"GdamSecondOrderFilter",
			sizeof(GdamSecondOrderFilter),
			sizeof(GdamSecondOrderFilterClass),
			(GtkClassInitFunc) gdam_second_order_filter_class_init,
			(GtkObjectInitFunc) gdam_second_order_filter_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_FILTER;
		parent_class = gtk_type_class(parent);
		second_order_filter_type = gtk_type_unique(parent,
		                                 &second_order_filter_info);
	}
	return second_order_filter_type;
}

