/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsequencerskin.h"
#include <gdam/component/gdambuffermanager.h>
#include "../gdamskinutil.h"
#include "../../component/plugins/gdammididevice.h"
#include <gtk/gtkdrawingarea.h>
#include <gtk/gtksignal.h>
#include <gtk/gtklabel.h>
#include <gtk/gtktogglebutton.h>
#include <gtk/gtkbutton.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>

#define gdam_skin_pop_message(skin, context)
#define gdam_skin_push_message(skin, context, string)

enum {
	SELECTION_MODE_SET,
	SELECTION_MODE_TOGGLE,
	SELECTION_MODE_ADD
};
static GtkObjectClass* parent_class = NULL;

typedef struct _GdamBufferInfo GdamBufferInfo;
struct _GdamBufferInfo {
	int		buffer_id;
	int		ref_count;
	char*		buffer_name;
	gdouble         hertz;
	int		render_offset_x, render_offset_y;
};
enum _SequencerDisplayMode {
	BY_GAIN,
	BY_RESAMPLE
};
static void delayed_redraw(GdamSequencerSkin* skin);
static void gdam_sequencer_skin_class_init(GdamSkinClass* skin_class);

#define MAX_GAIN                1.0

static gboolean       gdam_sequencer_build_new_model
                                                 (GdamSkin*      skin,
                                                  GdamChannel*   channel,
						  int		 num_options,
						  char**         options);

static int      equal_tempered_table_size;
static gdouble* log_equal_tempered_table = NULL;

static gdouble  min_log_hertz_global;
static gdouble  max_log_hertz_global;

static gdouble* log_equal_tempered_scale()
{
	int i;
	if (log_equal_tempered_table)
		return log_equal_tempered_table;
	equal_tempered_table_size = 12 * 12;
	log_equal_tempered_table = g_new(gdouble, equal_tempered_table_size);
	for (i = 0; i < equal_tempered_table_size; i++)
		log_equal_tempered_table[i] =
			log(440.0 * pow(2.0, (gdouble)(i - 48 - 9) / 12.0));
	min_log_hertz_global = log_equal_tempered_table[0];
	max_log_hertz_global = log_equal_tempered_table[i-1];
	return log_equal_tempered_table;
}

static void gdam_sequencer_skin_init(GdamSequencerSkin* sequencer_skin)
{
	sequencer_skin->default_resample = 1.0;
	sequencer_skin->default_gain = 0.3;
	sequencer_skin->circle_radius = 10;
	sequencer_skin->selected_elements = g_hash_table_new(NULL, NULL);
	sequencer_skin->selection_mode = SELECTION_MODE_SET;

	sequencer_skin->hidden_buffers = g_hash_table_new(NULL, NULL);

	sequencer_skin->log_resample_grid = log_equal_tempered_scale();
	sequencer_skin->num_resample_grid = equal_tempered_table_size;
	sequencer_skin->gain_grid = NULL;
	sequencer_skin->num_gain_grid = 0;
	sequencer_skin->max_log_hertz = log(8000);
	sequencer_skin->min_log_hertz = log(110);
}

GtkType gdam_sequencer_skin_get_type()
{
	static GtkType sequencer_skin_type = 0;
	if (!sequencer_skin_type) {
		static const GtkTypeInfo sequencer_skin_info = {
			"GdamSequencerSkin",
			sizeof(GdamSequencerSkin),
			sizeof(GdamSequencerSkinClass),
			(GtkClassInitFunc) gdam_sequencer_skin_class_init,
			(GtkObjectInitFunc) gdam_sequencer_skin_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_SKIN;
		parent_class = gtk_type_class(parent);
		sequencer_skin_type = gtk_type_unique(parent,
		                                 &sequencer_skin_info);
	}
	return sequencer_skin_type;
}


static GdamBufferInfo* get_buffer_info_from_id(GdamSequencerSkin* skin,  
                                               int id)
{
	GdamBufferInfo* info;
	if (skin->buffer_info_hash == NULL)
		skin->buffer_info_hash = g_hash_table_new(NULL, NULL);
	info = (GdamBufferInfo*) g_hash_table_lookup(skin->buffer_info_hash,
	                                             GINT_TO_POINTER(id));
	if (info != NULL) {
		info->ref_count++;
		return info;
	}


	info = g_new0(GdamBufferInfo, 1);
	info->buffer_id = id;
	info->buffer_name = g_strdup_printf("%d", id);
	info->ref_count = 1;
	info->render_offset_x = info->render_offset_y = -10000;
	info->hertz = 440.0;

	g_hash_table_insert(skin->buffer_info_hash, GINT_TO_POINTER(id), info);
	return info;
}

void                   deref_buffer           (GdamSequencerSkin* skin,
                                               int id)
{
	GdamBufferInfo* info;
	g_return_if_fail(skin->buffer_info_hash != NULL);
	info = g_hash_table_lookup(skin->buffer_info_hash, GINT_TO_POINTER(id));
	g_return_if_fail(info != NULL);
	g_return_if_fail(info->ref_count > 0);
	if (--(info->ref_count) == 0)  {
		g_hash_table_remove(skin->buffer_info_hash,
		                    GINT_TO_POINTER(id));
		if (info->buffer_name) g_free(info->buffer_name);
		g_free(info);
	}
}


/*
 *  ____                     _                  _
 * |  _ \ _ __ __ ___      _(_)_ __   __ _     / \   _ __ ___  __ _
 * | | | | '__/ _` \ \ /\ / / | '_ \ / _` |   / _ \ | '__/ _ \/ _` |
 * | |_| | | | (_| |\ V  V /| | | | | (_| |  / ___ \| | |  __/ (_| |
 * |____/|_|  \__,_| \_/\_/ |_|_| |_|\__, | /_/   \_\_|  \___|\__,_|
 *                                   |___/
 */

static void alloc_colors(GdamSequencerSkin* skin)
{
	int state = GTK_WIDGET_STATE(skin->drawing_area);
	if (skin->backing_pixmap == NULL)
		return;
	skin->circle_gc = skin->drawing_area->style->bg_gc[state];
	skin->background = skin->drawing_area->style->fg_gc[state];
	skin->text_gc = skin->background;
	skin->font = skin->drawing_area->style->font;
	skin->grid_gc = gdk_gc_new(skin->backing_pixmap);
	gdk_gc_set_foreground(skin->grid_gc,
	                      util_alloc_color_cached(65535, 65535, 0));
	skin->selection_gc = gdk_gc_new(skin->backing_pixmap);
	gdk_gc_set_foreground(skin->selection_gc,
	                      util_alloc_color_cached(0, 65535, 0));
	skin->circle_boundary_gc = gdk_gc_new(skin->backing_pixmap);
	gdk_gc_set_foreground(skin->circle_boundary_gc,
	                      util_alloc_color_cached(65535, 0, 0));
}

typedef struct _RenderInfo RenderInfo;
struct _RenderInfo {
	int x,y,w,h;
	GdamSequencerSkin* skin;
};

static gdouble get_sample_hertz(GdamSequencerSkin* skin,
                                GdamSequencerElement* element)
{
	GdamBufferInfo* info;
	int id = element->buffer_id;
	if (skin->buffer_info_hash == NULL)
		return 440.0;
	info = g_hash_table_lookup(skin->buffer_info_hash, GINT_TO_POINTER(id));
	if (info == NULL)
		return 440.0;
	return info->hertz;
}

static int get_y_position(GdamSequencerSkin*    skin,
                          GdamSequencerElement* element)
{
	switch (skin->display_mode) {
		case BY_GAIN:
			return skin->draw_height
			     - element->gain * skin->draw_height / MAX_GAIN;
		case BY_RESAMPLE: {
			gdouble log_hertz;
			gdouble sample_hertz = get_sample_hertz(skin, element);
			log_hertz = log(element->rate * sample_hertz);
			return skin->draw_height
			       - (log_hertz - skin->min_log_hertz)
			       * skin->draw_height
			       / (skin->max_log_hertz - skin->min_log_hertz);
		}
		default: g_assert_not_reached();
	}
	return 0;
}

static inline GdamSequencer* get_sequencer(GdamSequencerSkin* skin)
{
	return GDAM_SEQUENCER(GDAM_SKIN(skin)->model);
}

static void get_subsample_position(GdamSequencerElement* element, 
                                   GdamSequencerSkin*    skin,
                                   int*                  out_x,
				   int*                  out_y)
{
	gdouble num_beats = get_sequencer(skin)->num_beats;
	if (num_beats <= 0)
		*out_x = 0;
	else {
		gdouble where = element->beat_offset - skin->min_beat;
		where /= (skin->max_beat - skin->min_beat);
		*out_x = skin->draw_width * where;
	}
	*out_y = get_y_position(skin, element);
}

static void 
get_string_offset(GdkFont* font, const char* string,
                  int* offset_x, int* offset_y)
{
	int lbearing, rbearing, width, ascent, descent;
	gdk_string_extents(font, string,
	                   &lbearing, &rbearing,
			   &width, &ascent, &descent);
	*offset_x = (lbearing+rbearing) / 2;
	*offset_y = (descent - ascent) / 2;
}

static void render_subsample(GdamSequencerElement* subsample_info,
                             RenderInfo* render_info)
{
	GdkPixmap* pixmap = render_info->skin->backing_pixmap;
	int x, y, radius;
	GdamBufferInfo* buffer_info;
	GdamSequencerSkin* skin = render_info->skin;
	GdkGC* gc_inner; GdkGC* gc_outer;
	get_subsample_position(subsample_info, skin, &x, &y);
	radius = skin->circle_radius;

	if (skin->in_transit && g_hash_table_lookup(skin->in_transit,
	                                            subsample_info))
		return;

	if (g_hash_table_lookup(skin->hidden_buffers,
	                        GUINT_TO_POINTER(subsample_info->buffer_id)))
		return;

	/*
	 * Does the box around x/y touch the exposed area?
	 */
	if (x - radius > render_info->x + render_info->w
	 || x + radius < render_info->x
	 || y - radius > render_info->y + render_info->h
	 || y + radius < render_info->y)
	 	return;

	gc_inner = skin->circle_gc;
	gc_outer = skin->circle_boundary_gc;

	if (g_hash_table_lookup(skin->selected_elements, subsample_info)) {
		GdkGC* tmp = gc_inner;
		gc_inner = gc_outer;
		gc_outer = tmp;
	}

	gdk_draw_arc(pixmap,
		     gc_inner,
		     TRUE,
		     x - radius + 1,
		     y - radius + 1,
		     radius * 2 - 1,
		     radius * 2 - 1,
		     0,
		     360*64);
	gdk_draw_arc(pixmap,
		     gc_outer,
		     FALSE,
		     x - radius + 1,
		     y - radius + 1,
		     radius * 2 - 1,
		     radius * 2 - 1,
		     0,
		     360*64);

	buffer_info = get_buffer_info_from_id(skin, subsample_info->buffer_id);
	if (buffer_info->render_offset_x < -1000) {
		get_string_offset(skin->drawing_area->style->font,
		                  buffer_info->buffer_name,
				  &buffer_info->render_offset_x,
				  &buffer_info->render_offset_y);
	}

	gdk_draw_string(pixmap,
	                skin->font,
		        skin->text_gc,
		        x - buffer_info->render_offset_x,
		        y - buffer_info->render_offset_y,
		        buffer_info->buffer_name);
}

static void copy_backing_area(GdamSequencerSkin* skin, int x, int y, int w, int h)
{
        gdk_draw_pixmap(skin->drawing_area->window,
                        skin->drawing_area->style
				->fg_gc[GTK_WIDGET_STATE(skin->drawing_area)],
                        skin->backing_pixmap,
                        x, y, x, y, w, h);
}
static void redraw_area(GdamSequencerSkin* skin, int x, int y, int w, int h)
{
	RenderInfo render_info;
	render_info.x = x;
	render_info.y = y;
	render_info.w = w;
	render_info.h = h;
	render_info.skin = skin;
	if (!skin->circle_gc)
		alloc_colors(skin);
	if (skin->backing_pixmap == NULL)
		return;
	gdk_draw_rectangle(skin->backing_pixmap,
	                   skin->background,
		           TRUE,
		           x,y,w,h);
	if (get_sequencer(skin)->num_sub_beats && skin->grid_gc) {
		int grid_size = skin->num_beat_grid;
		int skip = get_sequencer(skin)->num_sub_beats;
		int i;
		for (i = 0; i < grid_size; i += skip) {
			gdouble t = skin->beat_grid[i];
			t -= skin->min_beat;
			t /= (skin->max_beat - skin->min_beat);
			if (0 <= t && t <= 1.0) {
				int pos = t * skin->draw_width;
				gdk_draw_line(skin->backing_pixmap,
				              skin->grid_gc,
					      pos, y, pos, y + h);
			}
		}
	}
	if (skin->grid_gc) {
		int grid_size;
		gdouble* grid;
		int i;
		gdouble top, bottom;
		if (skin->display_mode == BY_RESAMPLE) {
			grid = skin->log_resample_grid;
			grid_size = skin->num_resample_grid;
			top = skin->max_log_hertz;
			bottom = skin->min_log_hertz;
		} else {
			grid = skin->gain_grid;
			grid_size = skin->num_gain_grid;
			top = 2.0;
			bottom = 0.0;
		}
		for (i = 0; i < grid_size; i++) {
			gdouble t = grid[i];
			if (bottom <= t && t <= top) {
				int pos = (int)((t-bottom) * skin->draw_height
				                           / (top-bottom));
				/* reverse y to pixel-style coords */
				pos = skin->draw_height - 1 - pos;
				gdk_draw_line(skin->backing_pixmap,
				              skin->grid_gc,
					      x, pos, x + w, pos);
			}
		}
	}
	g_list_foreach(get_sequencer(skin)->elements,
	               (GFunc)render_subsample,
		       &render_info);
	copy_backing_area(skin, x, y, w, h);
}

#define DIST(a,b)  ( ((a) < (b)) ? ((b) - (a)) : ((a) - (b)) )

/* This function should use a binary search. */
static void match_line(gdouble* adjust, int num, const gdouble* vals)
{
	int closest;
	gdouble closest_dist;
	gdouble saught = *adjust;
	int i;
	if (num <= 0)
		return;
	g_return_if_fail(vals != NULL);

	closest = 0;
	closest_dist = DIST(saught, vals[0]);
	for (i = 1; i < num; i++) {
		gdouble dist = DIST(saught, vals[i]);
		if (dist < closest_dist) {
			closest = i;
			closest_dist = dist;
		}
	}
	*adjust = vals[closest];
}

static void
snap_to_grid(GdamSequencerSkin*      skin,
             int                     x,
	     int                     y,
	     GdamSequencerElement*   element)
{
	GdamSequencer* sequencer = get_sequencer(skin);
	gdouble start_beat = (gdouble)x * (skin->max_beat - skin->min_beat)
	                              / skin->draw_width + skin->min_beat;
	gdouble temp = start_beat;
	match_line(&temp, skin->num_beat_grid, skin->beat_grid);
	gdam_debug("snap_to_grid: snapped beat from %f to %f",start_beat,temp);
	element->beat_offset = temp;
	gdam_sequencer_update_sequence(sequencer);

	/* For the purposes of the following computations
	 * and anything involving resample/gain we invert
	 * y, so it corresponds to sheet music, math, etc,
	 * instead of computers. */
	y = skin->draw_height - y;
	switch (skin->display_mode) {
		case BY_GAIN: {
			gdouble gain;
			gain = y * MAX_GAIN / skin->draw_height;
			gdam_debug("snap_to(GAIN): y=%d, draw_height=%d; "
				"gain=%g", y,skin->draw_height,gain);
			match_line(&gain,
				   skin->num_gain_grid,
			           skin->gain_grid);
			element->gain = gain;
			break;
		}
		case BY_RESAMPLE: {
			gdouble log_hertz;
			log_hertz = (skin->max_log_hertz - skin->min_log_hertz)
			          * y / skin->draw_height
				  + skin->min_log_hertz;
			match_line(&log_hertz,
				   skin->num_resample_grid,
			           skin->log_resample_grid);
			element->rate = exp(log_hertz)
			              / get_sample_hertz(skin, element);
			break;
		}
		default: g_assert_not_reached();
	}
}

static gint
sequence_edit_expose(GtkDrawingArea* area,
                     GdkEventExpose* expose_event,
		     GdamSequencerSkin* skin)
{
	redraw_area(skin, 
		    expose_event->area.x, expose_event->area.y,
                    expose_event->area.width, expose_event->area.height);
	return 0;
}

GList* find_submodel_list_el(GdamSequencerSkin* skin, int x, int y)
{
	GList* at;
	GList* found = NULL;
	for (at = get_sequencer(skin)->elements; at; at = at->next) {
		int pos_x, pos_y;
		GdamSequencerElement* subsample;
		gboolean is_selected;

		is_selected = g_hash_table_lookup(skin->selected_elements,
		                                  at->data) != NULL;
		if (found != NULL && !is_selected)
			continue;

		subsample = (GdamSequencerElement*)at->data;
		get_subsample_position(subsample,
		                       skin,
                                       &pos_x, &pos_y);
		if (pos_x - skin->circle_radius <= x
		 && x <= pos_x + skin->circle_radius
		 && pos_y - skin->circle_radius <= y
		 && y <= pos_y + skin->circle_radius) {
		 	if (is_selected)
				return at;
			found = at;
		}
	}
	return found;
}
GdamSequencerElement* find_submodel(GdamSequencerSkin* skin, int x, int y)
{
	GList* out = find_submodel_list_el(skin, x, y);
	return (GdamSequencerElement*) (out ? out->data : NULL);
}


static void delete_subinfo(GdamSequencerSkin*    skin,
                           GdamSequencerElement* element)
{
	int sample_x, sample_y, radius = skin->circle_radius;

	deref_buffer(skin, element->buffer_id);

	g_hash_table_remove(skin->selected_elements, element);

	get_subsample_position(element, skin, &sample_x, &sample_y);
	gdam_sequencer_remove_element(get_sequencer(skin), element);
	redraw_area(skin, sample_x - radius - 1, sample_y - radius - 1,
	                  2 * radius + 3, 2 * radius + 3);
}

static void add_element_ptr(gpointer el, gpointer unused, gpointer arr)
{
	GdamSequencerElement** element = *(GdamSequencerElement***)arr;
	*element++ = (GdamSequencerElement*)el;
	*(GdamSequencerElement***)arr = element;
}

static void delete_selected(GdamSequencerSkin*    skin)
{
	GdamSequencerElement** elements_to_kill;
	GdamSequencerElement** tmp;
	GdamSequencer* sequencer = get_sequencer(skin);
	int n = g_hash_table_size(skin->selected_elements);
	int i;
	tmp = (GdamSequencerElement**)alloca(sizeof(GdamSequencerElement*) * n);
	elements_to_kill = tmp;
	g_hash_table_foreach(skin->selected_elements, add_element_ptr, &tmp);
	g_assert(elements_to_kill + n == tmp);
	gdam_sequencer_freeze(sequencer);
	for (i = 0; i < n; i++)
		delete_subinfo(skin, elements_to_kill[i]);
	gdam_sequencer_thaw(sequencer);
}


/*
 *  ____       _           _   _
 * / ___|  ___| | ___  ___| |_(_) ___  _ __     __ _ _ __ ___  __ _
 * \___ \ / _ \ |/ _ \/ __| __| |/ _ \| '_ \   / _` | '__/ _ \/ _` |
 *  ___) |  __/ |  __/ (__| |_| | (_) | | | | | (_| | | |  __/ (_| |
 * |____/ \___|_|\___|\___|\__|_|\___/|_| |_|  \__,_|_|  \___|\__,_|
 * 
 */
static void unselect_all_elements(GdamSequencerSkin* skin)
{
	g_hash_table_destroy(skin->selected_elements);
	skin->selected_elements = g_hash_table_new(NULL, NULL);
}

static void begin_selection(GdamSequencerSkin* skin, int x, int y)
{
	skin->has_selection = TRUE;
	skin->selection[0].x = skin->selection[1].x = x;
	skin->selection[0].y = skin->selection[1].y = y;
}

static void draw_selection(GdamSequencerSkin* skin)
{
	GtkWidget* widget = skin->drawing_area;
	int x1 = skin->selection[0].x;
	int y1 = skin->selection[0].y;
	int x2 = skin->selection[1].x;
	int y2 = skin->selection[1].y;
	gdk_draw_line(widget->window,
		      skin->selection_gc,
		      x1, y1, x1, y2);
	gdk_draw_line(widget->window,
		      skin->selection_gc,
		      x2, y1, x2, y2);
	gdk_draw_line(widget->window,
		      skin->selection_gc,
		      x1, y1, x2, y1);
	gdk_draw_line(widget->window,
		      skin->selection_gc,
		      x1, y2, x2, y2);
}

static void get_selection_rect(GdamSequencerSkin* skin, GdkRectangle* rect)
{
	int width, height;
	width = skin->selection[1].x - skin->selection[0].x;
	height = skin->selection[1].y - skin->selection[0].y;
	if (width < 0) {
		rect->width = -width + 1;
		rect->x = skin->selection[1].x;
	} else {
		rect->width = width + 1;
		rect->x = skin->selection[0].x;
	}
	if (height < 0) {
		rect->height = -height + 1;
		rect->y = skin->selection[1].y;
	} else {
		rect->height = height + 1;
		rect->y = skin->selection[0].y;
	}
}
static void undraw_selection(GdamSequencerSkin* skin)
{
	GdkRectangle rect;
	get_selection_rect(skin, &rect);

	copy_backing_area(skin, rect.x, rect.y, 1, rect.height);
	copy_backing_area(skin,
	                  rect.x + rect.width - 1,
	                  rect.y,
			  1,
			  rect.height);
	if (rect.width > 2) {
		copy_backing_area(skin,
				  rect.x+1, rect.y,
				  rect.width-2, 1);
		copy_backing_area(skin,
				  rect.x+1, rect.y + rect.height - 1,
				  rect.width-2, 1);
	}
}
static void update_selection(GdamSequencerSkin* skin, int x, int y)
{
	undraw_selection(skin);
	skin->selection[1].x = x;
	skin->selection[1].y = y;
	draw_selection(skin);
}

static gboolean find_union(gpointer key, gpointer value, gpointer h)
{
	g_hash_table_insert((GHashTable*)h, key, value);
	return TRUE;
}

static void handle_selected(GdamSequencerSkin*    skin,
                            GdamSequencerElement* element)
{
	GHashTable* selected = skin->selected_elements;
	int mode = skin->selection_mode;
	if (mode == SELECTION_MODE_SET
	 || mode == SELECTION_MODE_ADD) {
		g_hash_table_insert(selected, element, element);
	} else { /* SELECTION_MODE_TOGGLE */
		if (g_hash_table_lookup(selected, element))
			g_hash_table_remove(selected, element);
		else
			g_hash_table_insert(selected, element, element);
	}
}
static void finish_selection(GdamSequencerSkin* skin, int x, int y)
{
	GHashTable* redraw_table = g_hash_table_new(NULL, NULL);
	GList* elements = get_sequencer(skin)->elements;
	GdkRectangle rect;
	int radius = skin->circle_radius;
	int old_size = g_hash_table_size(skin->selected_elements);
	int new_size;
	undraw_selection(skin);
	skin->has_selection = FALSE;
	if (skin->selection_mode == SELECTION_MODE_SET)
		g_hash_table_foreach_remove(skin->selected_elements,
		                            find_union,
					    redraw_table);
	get_selection_rect(skin, &rect);
	while (elements != NULL) {
		GdamSequencerElement* el = elements->data;
		int sample_x, sample_y;
		get_subsample_position(el, skin, &sample_x, &sample_y);
		if (rect.x <= sample_x + radius
		 && sample_x - radius < rect.x + rect.width
		 && rect.y <= sample_y + radius
		 && sample_y - radius < rect.y + rect.height) {
		 	g_hash_table_insert(redraw_table, el, el);
			handle_selected(skin, el);
		}
		elements = elements->next;
	}

	/* TODO: incremental redraw using redraw_table here.... */
	//g_hash_table_foreach(redraw_area, redraw_one_func, skin);

	g_hash_table_destroy(redraw_table);
	new_size = g_hash_table_size(skin->selected_elements);

	{ char buf[1024];
	  gdam_skin_pop_message(GDAM_SKIN(skin),
	                        skin->selection_status_context);
	  snprintf(buf, sizeof(buf), "Had selected %d elements; "
	  			     "now %d elements are selected.",
				     old_size,
				     new_size); 
	  buf[sizeof(buf) - 1] = 0;
	  gdam_skin_push_message(GDAM_SKIN(skin),
	                         skin->selection_status_context,
				 buf); }

	delayed_redraw(skin);
}

/*
 *  ____                        _
 * |  _ \ _ __ __ _  __ _  __ _(_)_ __   __ _
 * | | | | '__/ _` |/ _` |/ _` | | '_ \ / _` |
 * | |_| | | | (_| | (_| | (_| | | | | | (_| |
 * |____/|_|  \__,_|\__, |\__, |_|_| |_|\__, |
 *                  |___/ |___/         |___/
 */
static void
draw_one_dragged_sample(GdamSequencerSkin* skin, GdamSequencerElement* element)
{
	int x,y;
	int radius = skin->circle_radius;
	get_subsample_position(element, skin, &x, &y);
	x += skin->drag_current_x - skin->drag_start_x;
	y += skin->drag_current_y - skin->drag_start_y;
	gdk_draw_arc(skin->drawing_area->window,
		     skin->circle_gc,
		     FALSE,
		     x - radius + 1,
		     y - radius + 1,
		     radius * 2 - 1,
		     radius * 2 - 1,
		     0,
		     360*64);
}
static void draw_dragged_samples(GdamSequencerSkin* skin)
{
	int i;
	for (i = 0; i < skin->num_dragging; i++)
		draw_one_dragged_sample(skin, skin->dragging + i);
}
static void
undraw_one_dragged_sample(GdamSequencerSkin* skin,
                          GdamSequencerElement* element)
{
	int x,y;
	int radius = skin->circle_radius;
	get_subsample_position(element, skin, &x, &y);
	x += skin->drag_current_x - skin->drag_start_x;
	y += skin->drag_current_y - skin->drag_start_y;
	copy_backing_area(skin,
	                  x - radius - 1,
			  y - radius - 1,
	                  2 * radius + 2,
			  2 * radius + 2);
}
static void undraw_dragged_samples(GdamSequencerSkin* skin)
{
	int i;
	for (i = 0; i < skin->num_dragging; i++)
		undraw_one_dragged_sample(skin, skin->dragging + i);
}

static void copy_element_to_array(gpointer key, gpointer val, gpointer udata)
{
	GdamSequencerElement* element = *(GdamSequencerElement**)udata;
	*element++ = *(GdamSequencerElement*)key;
	*(GdamSequencerElement**)udata = element;
}
static void copy_hash_element(gpointer key, gpointer val, gpointer user_data)
{
	g_hash_table_insert((GHashTable*)user_data, key, val);
}

static void
begin_dragging(GdamSequencerSkin*    skin,
               GdamSequencerElement* sample_info,
	       guint                 mask)
{
	gboolean copy = (mask & GDK_SHIFT_MASK) == GDK_SHIFT_MASK;

	g_free(skin->dragging);
	if (skin->in_transit != NULL) {
		g_hash_table_destroy(skin->in_transit);
		skin->in_transit = NULL;
	}

	/* Set up:
	 *      skin->in_transit
	 *      skin->num_dragging
	 *      skin->dragging. */
	if (g_hash_table_lookup(skin->selected_elements, sample_info)) {
		int num_dragging = g_hash_table_size(skin->selected_elements);
		GdamSequencerElement* tmp;
		skin->num_dragging = num_dragging;
		skin->dragging = g_new(GdamSequencerElement, num_dragging);
		tmp = skin->dragging;
		g_hash_table_foreach(skin->selected_elements,
		                     copy_element_to_array,
				     &tmp);
		if (!copy) {
			skin->in_transit = g_hash_table_new(NULL, NULL);
			g_hash_table_foreach(skin->selected_elements,
		                             copy_hash_element,
				             skin->in_transit);
		}
		skin->was_dragging_selected = TRUE;
	} else {
		skin->num_dragging = 1;
		skin->dragging = g_new(GdamSequencerElement, 1);
		*skin->dragging = *sample_info;
		if (!copy) {
			skin->in_transit = g_hash_table_new(NULL, NULL);
			g_hash_table_insert(skin->in_transit,
			                    sample_info,
					    sample_info);
		}
		skin->was_dragging_selected = FALSE;
	}

	/* TODO: should only redraw the in_transit area. */
	redraw_area(skin, 0, 0, skin->draw_width, skin->draw_height);
	draw_dragged_samples(skin);
}
static void remove_element(gpointer el, gpointer unused, gpointer seq)
{
	gdam_sequencer_remove_element(GDAM_SEQUENCER(seq),
	                              (GdamSequencerElement*) el);
}

static void finished_dragging(GdamSequencerSkin* skin)
{
	int num_dragging = skin->num_dragging;
	GdamSequencer* sequencer = get_sequencer(skin);
	int i;
	int move_x, move_y;
	gdouble delta_start;
	gdouble delta_gain, hertz_scale = 1.0, frac_y, delta_log_hertz;
	move_x = skin->drag_current_x - skin->drag_start_x;
	move_y = skin->drag_current_y - skin->drag_start_y;

	delta_start = sequencer->num_beats * move_x / skin->draw_width;
	frac_y = (gdouble) -move_y / skin->draw_height;
	if (skin->display_mode == BY_GAIN) {
		delta_gain = MAX_GAIN * frac_y;
		delta_log_hertz = 0.0;
	} else {
		delta_gain = 0.0;
		delta_log_hertz = (skin->max_log_hertz - skin->min_log_hertz)
		               * frac_y;
	}

	/* TODO: make this snapping optional. */
	if (sequencer->num_sub_beats) {
		delta_start *= sequencer->num_sub_beats;
		delta_start += 0.5;
		delta_start = floor(delta_start);
		delta_start /= sequencer->num_sub_beats;
	}

	/* HACK: snap to equal tempered scale. */
	{
		gdouble incr = log(pow(2.0, 1.0/12.0));
		delta_log_hertz += incr / 2.0;
		delta_log_hertz /= incr;
		delta_log_hertz = (int)delta_log_hertz;
		delta_log_hertz *= incr;
		hertz_scale = exp(delta_log_hertz);
	}

	gdam_sequencer_freeze(sequencer);
	unselect_all_elements(skin);
	for (i = 0; i < num_dragging; i++) {
		GdamSequencerElement* element = skin->dragging + i;
		//int id = element->buffer_id;
		element = gdam_sequencer_add_element(sequencer,
	                        element->buffer_id,
				element->beat_offset + delta_start,
		                element->rate * hertz_scale,
				element->gain + delta_gain,
                                element->sample_offset,
				element->output_length);
		if (skin->was_dragging_selected)
			g_hash_table_insert(skin->selected_elements,
			                    element, element);
	}
	if (skin->in_transit) {
		g_hash_table_foreach(skin->in_transit,
		                     remove_element,
				     sequencer);
		g_hash_table_destroy(skin->in_transit);
		skin->in_transit = NULL;
	}

	g_free(skin->dragging);
	skin->dragging = NULL;
	skin->num_dragging = 0;

	gdam_sequencer_thaw(sequencer);
	delayed_redraw(skin);
}

/*  _              _   _          __  __         
 * | |___  __ __ _| | | |__ _  _ / _|/ _|___ _ _ 
 * | / _ \/ _/ _` | | | '_ \ || |  _|  _/ -_) '_|
 * |_\___/\__\__,_|_| |_.__/\_,_|_| |_| \___|_|  
 *
 *                                              _   
 *  _ __  __ _ _ _  __ _ __ _ ___ _ __  ___ _ _| |_ 
 * | '  \/ _` | ' \/ _` / _` / -_) '  \/ -_) ' \  _|
 * |_|_|_\__,_|_||_\__,_\__, \___|_|_|_\___|_||_\__|
 *                      |___/                       
 */
typedef struct _GdamSequencerBufferInfo GdamSequencerBufferInfo;

struct _GdamSequencerBufferInfo {
	guint			buffer_id;
	GdamSequencerSkin*	sequencer_skin;
	GtkWidget*		topwidget;
};

static void select_buffer_by_info     (GdamSequencerBufferInfo* info)
{
	GList* el;
	GdamSequencer* sequencer;
	GHashTable* selected_elements;
	GdamSequencerSkin* sequencer_skin = info->sequencer_skin;
	sequencer = get_sequencer(sequencer_skin);
	unselect_all_elements(sequencer_skin);
	selected_elements = sequencer_skin->selected_elements;
	for (el = sequencer->elements; el != NULL; el = el->next) {
		GdamSequencerElement* element = el->data;
		if (element->buffer_id == info->buffer_id)
			g_hash_table_insert(selected_elements,
					    element,
					    element);
	}
	delayed_redraw(sequencer_skin);
}

static void toggle_hide_buffer_by_info(GtkToggleButton*         toggle,
                                       GdamSequencerBufferInfo* info)
{
	gboolean hidden = GTK_TOGGLE_BUTTON(toggle)->active;
	GHashTable* hidden_buffers = info->sequencer_skin->hidden_buffers;
	gpointer buffer_id = GUINT_TO_POINTER(info->buffer_id);
	if (hidden)
		g_hash_table_insert(hidden_buffers, buffer_id, buffer_id);
	else
		g_hash_table_remove(hidden_buffers, buffer_id);
	delayed_redraw(info->sequencer_skin);
}

static void delete_one_buffer(gpointer el, gpointer skin)
{
	delete_subinfo(GDAM_SEQUENCER_SKIN(skin), (GdamSequencerElement*) el);
}

static void delete_buffer_by_info(GdamSequencerBufferInfo* info)
{
	GdamSequencerSkin* skin = info->sequencer_skin;
	GdamSequencer* sequencer = get_sequencer(skin);
	GList* at;
	GSList* to_delete = NULL;
	
	/* delete the buffer id from the list of sequencer elements. */
	for (at = sequencer->elements; at != NULL; at = at->next) {
		GdamSequencerElement* element;
		element = at->data;
		if (element->buffer_id == info->buffer_id) {
			to_delete = g_slist_prepend(to_delete, element);
		}
	}

	g_slist_foreach(to_delete, delete_one_buffer, skin);
	g_slist_free(to_delete);

	g_hash_table_remove(skin->hidden_buffers,
			    GUINT_TO_POINTER(info->buffer_id));

	gtk_widget_destroy(info->topwidget);
}

static void on_buffer_topwidget_destroy(GdamSequencerBufferInfo* info)
{
	GdamSequencerSkin* skin = info->sequencer_skin;
	g_hash_table_remove(skin->buffer_id_to_info,
			    GUINT_TO_POINTER(info->buffer_id));
	g_free(info);
}

static void maybe_track_buffer_id(GdamSequencerSkin* skin,
		                  guint              buffer_id)
{
	char label_buf[255];
	GdamSequencerBufferInfo* info;
	GtkWidget* new_widget;
	GtkWidget* sub_widget;

	if (skin->buffer_id_holder == NULL)
		return;

	if (skin->buffer_id_to_info == NULL) {
		skin->buffer_id_to_info = g_hash_table_new(NULL, NULL);
	}

	if (g_hash_table_lookup(skin->buffer_id_to_info,
				GUINT_TO_POINTER(buffer_id))) {
		return;
	}

	info = g_new(GdamSequencerBufferInfo, 1);

	info->buffer_id = buffer_id;
	info->sequencer_skin = skin;
	info->topwidget = new_widget = gtk_hbox_new(FALSE, 0);

	snprintf(label_buf, sizeof(label_buf), "Buffer %d", buffer_id);
	sub_widget = gtk_label_new(label_buf);
	gtk_widget_show(sub_widget);
	gtk_box_pack_start_defaults(GTK_BOX(new_widget), sub_widget);

	sub_widget = gtk_button_new_with_label("Select");
	gtk_signal_connect_object(GTK_OBJECT(sub_widget),
			          "clicked",
				  GTK_SIGNAL_FUNC(select_buffer_by_info),
				  (GtkObject*) info);
	gtk_widget_show(sub_widget);
	gtk_box_pack_start_defaults(GTK_BOX(new_widget), sub_widget);

	sub_widget = gtk_toggle_button_new_with_label("Hide");
	gtk_signal_connect(GTK_OBJECT(sub_widget),
			   "toggled",
		           GTK_SIGNAL_FUNC(toggle_hide_buffer_by_info),
			   info);
	gtk_widget_show(sub_widget);
	gtk_box_pack_start_defaults(GTK_BOX(new_widget), sub_widget);

	sub_widget = gtk_button_new_with_label("Delete");
	gtk_signal_connect_object(GTK_OBJECT(sub_widget),
			          "clicked",
				  GTK_SIGNAL_FUNC(delete_buffer_by_info),
				  (GtkObject*) info);
	gtk_widget_show(sub_widget);
	gtk_box_pack_start_defaults(GTK_BOX(new_widget), sub_widget);

	gtk_signal_connect_object(GTK_OBJECT(new_widget),
			          "destroy",
				  GTK_SIGNAL_FUNC(on_buffer_topwidget_destroy),
				  (GtkObject*) info);

	gtk_widget_show(new_widget);
	gtk_container_add(skin->buffer_id_holder, new_widget);

	g_hash_table_insert(skin->buffer_id_to_info,
		            GUINT_TO_POINTER(buffer_id),
			    info);
}

/*
 *   ____     _ _    _____                 _
 *  / ___| __| | | _| ____|_   _____ _ __ | |_ ___
 * | |  _ / _` | |/ /  _| \ \ / / _ \ '_ \| __/ __|
 * | |_| | (_| |   <| |___ \ V /  __/ | | | |_\__ \
 *  \____|\__,_|_|\_\_____| \_/ \___|_| |_|\__|___/
 * 
 */
static gint
sequence_edit_button_press(GtkDrawingArea* area,
                           GdkEventButton* button_event,
		           GdamSequencerSkin* skin)
{
	int press_x = button_event->x;
	int press_y = button_event->y;
	int width, height;
	GdamSequencerElement* sample_info;
	int buffer_id;
	gdouble beat;
	GdamSequencer* sequencer = get_sequencer(skin);
	gdam_debug("sequence_edit_button_press");
	g_return_val_if_fail(skin->dragging == NULL, 0);
	if (button_event->button == 2) {
		int state = button_event->state;
		gboolean is_shifted, is_ctrl;
		begin_selection(skin, press_x, press_y);
		is_ctrl = ((state&GDK_CONTROL_MASK) == GDK_CONTROL_MASK);
		is_shifted = ((state&GDK_SHIFT_MASK) == GDK_SHIFT_MASK);
		gdam_skin_pop_message(GDAM_SKIN(skin),
				      skin->selection_context_id);
		if (is_shifted) {
			gdam_skin_push_message(GDAM_SKIN(skin),
				skin->selection_context_id,
				"The framed items will be added to "
				"the selection set.");
			skin->selection_mode = SELECTION_MODE_ADD;
		} else if (is_ctrl) {
			gdam_skin_push_message(GDAM_SKIN(skin),
				skin->selection_context_id,
				"The framed items will be toggled between "
				"selected and unselected.");
			skin->selection_mode = SELECTION_MODE_TOGGLE;
		} else {
			gdam_skin_push_message(GDAM_SKIN(skin),
				skin->selection_context_id,
				"The framed items will be selected; the "
				"current selection will be cleared.");
			skin->selection_mode = SELECTION_MODE_SET;
		} return TRUE;
	}
	if ((sample_info=find_submodel(skin,
	                                   press_x, press_y)) != NULL) {
		int pos_x, pos_y;
		if (button_event->button == 3) {
			if (g_hash_table_lookup(skin->selected_elements,
			                        sample_info))
				delete_selected(skin);
			else
				delete_subinfo(skin, sample_info);
			return TRUE;
		}
		get_subsample_position(sample_info, skin, &pos_x, &pos_y);
		skin->drag_start_x = press_x;
		skin->drag_start_y = press_y;
		begin_dragging(skin, sample_info, button_event->state);
		return TRUE;
	}

	if (button_event->button == 3)
		return 0;

	buffer_id = gdam_buffer_manager_obtain_id(GDAM_MODEL(sequencer));
	if (buffer_id == 0) {
		g_message("no sample has been chosen -- cannot add");
		return 0;
	}


	gdam_sequencer_freeze(sequencer);

	width = skin->draw_width;
	height = skin->draw_height;
	  /* TODO: are we sure this position takes into account that we might
	   * be viewing only part of the sequence? */
	beat = press_x * sequencer->num_beats / width + skin->min_beat;
	sample_info = gdam_sequencer_add_element(
				sequencer,
	                        buffer_id,
				beat,
		                skin->default_resample,
                                skin->default_gain,
	                        0,
				-1);
	gdam_buffer_manager_unref(buffer_id, (GdamModel*)sequencer);
	maybe_track_buffer_id(skin, buffer_id);
	snap_to_grid(skin, press_x, press_y, sample_info);
	get_subsample_position(sample_info, skin, &press_x, &press_y);
	redraw_area(skin,
	            press_x - skin->circle_radius,
	            press_y - skin->circle_radius,
	            skin->circle_radius * 2 + 1,
	            skin->circle_radius * 2 + 1);

	gdam_sequencer_thaw(sequencer);
	return TRUE;
}

static gint
sequence_edit_button_motion(GtkDrawingArea* area,
                            GdkEventMotion* button_event,
		            GdamSequencerSkin* skin)
{
	int x,y;
	guint state;
	if (button_event->is_hint)
		gdk_window_get_pointer(button_event->window, &x, &y, &state);
	else {
		x = button_event->x;
		y = button_event->y;
		state = button_event->state;
	}

	if (skin->has_selection) {
		update_selection(skin, button_event->x, button_event->y);
		return TRUE;
	}
	if (skin->dragging) {
		undraw_dragged_samples(skin);
		skin->drag_current_x = button_event->x;
		skin->drag_current_y = button_event->y;
		draw_dragged_samples(skin);
	}
	return TRUE;
}

static gint
sequence_edit_button_release(GtkDrawingArea* area,
                             GdkEventButton* button_event,
		             GdamSequencerSkin* skin)
{
	if (skin->has_selection) {
		finish_selection(skin, button_event->x, button_event->y);
		return 0;
	}
	if (!skin->dragging)
		return 0;

	undraw_dragged_samples(skin);

	skin->drag_current_x = button_event->x;
	skin->drag_current_y = button_event->y;

	finished_dragging(skin);
	return TRUE;
}

static gint
sequence_edit_configure(GtkWidget* widget,
                        GdkEventConfigure* configure_event,
		        GdamSequencerSkin* skin)
{
	if (skin->backing_pixmap)
		gdk_pixmap_unref(skin->backing_pixmap);
	skin->backing_pixmap = gdk_pixmap_new(widget->window,
	                              widget->allocation.width,
	                              widget->allocation.height,
				      -1);
	skin->draw_width = widget->allocation.width;
	skin->draw_height = widget->allocation.height;
	return FALSE;
}

/* XXX: these should be connectable... */
static void sort_by_resample(GdamSequencerSkin* skin)
{
	skin->display_mode = BY_RESAMPLE;
	redraw_area(skin, 0, 0, skin->draw_width, skin->draw_height);
}
static void sort_by_gain(GdamSequencerSkin* skin)
{
	skin->display_mode = BY_GAIN;
	redraw_area(skin, 0, 0, skin->draw_width, skin->draw_height);
}


static void connect_drawing_area(GtkDrawingArea*    drawing_area,
                                 GdamSequencerSkin* skin)
{
	skin->drawing_area = GTK_WIDGET(drawing_area);
	//gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), 800, 600);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "button_press_event",
			   GTK_SIGNAL_FUNC(sequence_edit_button_press),
			   skin);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "motion_notify_event",
			   GTK_SIGNAL_FUNC(sequence_edit_button_motion),
			   skin);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "button_release_event",
			   GTK_SIGNAL_FUNC(sequence_edit_button_release),
			   skin);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "expose_event",
			   GTK_SIGNAL_FUNC(sequence_edit_expose),
			   skin);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "configure_event",
			   GTK_SIGNAL_FUNC(sequence_edit_configure),
			   skin);
	gtk_widget_add_events(GTK_WIDGET(drawing_area),
	                                    GDK_EXPOSURE_MASK
                                          | GDK_BUTTON_PRESS_MASK
                                          | GDK_BUTTON_RELEASE_MASK
                                          | GDK_POINTER_MOTION_MASK
                                          | GDK_POINTER_MOTION_HINT_MASK);
	//gtk_widget_show(GTK_WIDGET(drawing_area));
}

/*
 *  __  __ _              _ _
 * |  \/  (_)___  ___ ___| | | __ _ _ __   ___  ___  _   _ ___
 * | |\/| | / __|/ __/ _ \ | |/ _` | '_ \ / _ \/ _ \| | | / __|
 * | |  | | \__ \ (_|  __/ | | (_| | | | |  __/ (_) | |_| \__ \
 * |_|  |_|_|___/\___\___|_|_|\__,_|_| |_|\___|\___/ \__,_|___/
 * 
 * 
 * __      ___ __ __ _ _ __  _ __   ___ _ __ ___
 * \ \ /\ / / '__/ _` | '_ \| '_ \ / _ \ '__/ __|
 *  \ V  V /| | | (_| | |_) | |_) |  __/ |  \__ \
 *   \_/\_/ |_|  \__,_| .__/| .__/ \___|_|  |___/
 *                    |_|   |_|
 */

#if 0
static GdamSequencerSkin* new_blank_skin(GdamSequencer* sequencer)
{

	GdamSequencerSkin* skin = g_new0(GdamSequencerSkin, 1);
	GDAM_SKIN(skin)->model = GDAM_MODEL(sequencer);
	skin->default_resample = 1.0;
	skin->default_gain = 1.0;
	skin->circle_radius = 10;
	return skin;
}
#endif

static gboolean is_blank(const char* string)
{
	while (*string && isspace(*string))
		string++;
	return *string == 0 ? TRUE : FALSE;
}

static void do_load(GdamSequencerSkin* skin)
{
	const char* filename;
	g_return_if_fail(skin->filename_entry != NULL);
	unselect_all_elements(skin);
	filename = gtk_entry_get_text(skin->filename_entry);
	g_return_if_fail(filename != NULL && !is_blank(filename));
	gdam_sequencer_load(get_sequencer(skin), filename);
	delayed_redraw(skin);
}

static void on_note_on(GdamMidiDevice* midi, GdamMidiNote* note, 
		       GdamSequencerSkin* skin)
{
	GdamSequencer* sequencer = GDAM_SEQUENCER(GDAM_SKIN(skin)->model);
	int buffer_id;
	double resample, gain;
	gboolean must_unref = FALSE;

	if (skin->local_buffers[0] != 0)
		buffer_id = skin->local_buffers[0];
	else {
		buffer_id = gdam_buffer_manager_obtain_id(
						GDAM_MODEL(sequencer));
		must_unref = TRUE;
	}

	if (buffer_id == 0) {
		g_message("no sample has been chosen -- cannot add from midi");
		return;
	}

	resample = (double)note->hertz / 440.0;
	gain = (double)note->velocity / 128.0;


	gdam_sequencer_add_element_immediate(sequencer,
					     buffer_id,
					     resample,
					     gain,
					     0,
					     -1);
	if (must_unref)
		gdam_buffer_manager_unref(buffer_id, (GdamModel*)sequencer);
	maybe_track_buffer_id(skin, buffer_id);
	delayed_redraw(skin);
}

static void do_midi(GdamSequencerSkin* skin){
	GdamMidiDevice* midi;
	midi = gdam_midi_device_get_unique("/dev/midi");
	if (midi == NULL) { 
		g_warning("Couldn't open /dev/midi");
		return;
	}
	gtk_signal_connect_while_alive(GTK_OBJECT(midi),
			   "note_on",
			   on_note_on,
			   skin,
			   (GtkObject*)skin);
}

static void do_save(GdamSequencerSkin* skin)
{
	const char* filename;
	g_return_if_fail(skin->filename_entry != NULL);
	filename = gtk_entry_get_text(skin->filename_entry);
	g_return_if_fail(filename != NULL && !is_blank(filename));
	gdam_sequencer_save(get_sequencer(skin), filename);
}


/*
 *  ____                   _                ___
 * |  _ \ __ _ _ __  _ __ (_)_ __   __ _   ( _ )
 * | |_) / _` | '_ \| '_ \| | '_ \ / _` |  / _ \/\
 * |  __/ (_| | | | | | | | | | | | (_| | | (_>  <
 * |_|   \__,_|_| |_|_| |_|_|_| |_|\__, |  \___/\/
 *                                 |___/
 *  _____                     _
 * |__  /___   ___  _ __ ___ (_)_ __   __ _
 *   / // _ \ / _ \| '_ ` _ \| | '_ \ / _` |
 *  / /| (_) | (_) | | | | | | | | | | (_| |
 * /____\___/ \___/|_| |_| |_|_|_| |_|\__, |
 *                                    |___/
 */
static void horizontal_zoom(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
	skin->max_beat = skin->min_beat + adj->value;
	if (skin->time_pan_adj) {
		skin->time_pan_adj->page_size = adj->value;
		gtk_adjustment_changed(adj);
	}
	delayed_redraw(skin);
}

static void horizontal_pan(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
	skin->max_beat += (adj->value - skin->min_beat);
	skin->min_beat = adj->value;
	delayed_redraw(skin);
}

static void vertical_zoom(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
	skin->max_log_hertz = skin->min_log_hertz + adj->value;
	if (skin->vertical_pan_adj) {
		skin->vertical_pan_adj->page_size = adj->value;
		gtk_adjustment_changed(adj);
	}
	delayed_redraw(skin);
}

static void vertical_pan(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
	skin->max_log_hertz += (adj->value - skin->min_log_hertz);
	skin->min_log_hertz = adj->value;
	delayed_redraw(skin);
}

/* 
 *  ____  _    _         _____                 _   _                 
 * / ___|| | _(_)_ __   |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
 * \___ \| |/ / | '_ \  | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 *  ___) |   <| | | | | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |____/|_|\_\_|_| |_| |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 */

typedef struct _IntInfo {
	GdamSequencerSkin* skin;
	int num;
} IntInfo;

static void on_beatnum_adjusted(GtkAdjustment* adj, GdamSequencerSkin* skin);
static void on_tempo_adjusted(GtkAdjustment* adj, gpointer user_data);
static void on_sequencer_set_beat_info(GtkAdjustment* adj, 
				       const GdamBeatInfo* info, 
				       gpointer user_data);
static void on_tempo_string_drop(const char* tempo, gpointer user_data);
static void on_buffer_id_drop(const char* id_string, gpointer user_data);
static void on_subbeats_adjusted(GtkAdjustment* adj, GdamSequencerSkin* skin);
static void on_asynch_step_size_adjusted(GtkAdjustment* adj, 
					 GdamSequencerSkin* skin);
static void on_lag_adjusted(GtkAdjustment* adj, GdamSequencerSkin* skin);

static void do_select_local_buffer(IntInfo* info)
{
	//GdamSequencerSkin* skin = info->skin;
	//GdamSequencer* sequencer = GDAM_SEQUENCER(GDAM_SKIN(skin)->model);
}

static void do_step_asynch(GdamSequencerSkin* skin){
	GdamSequencer* sequencer = get_sequencer(skin);
	gdam_debug("autostepping.");
	skin->asynch_position += skin->asynch_step_size;
	if (skin->asynch_position >= sequencer->num_beats)
		skin->asynch_position -= sequencer->num_beats;
	if (skin->asynch_position_label){
		char txt[16];
		sprintf(txt, "%3.2f", skin->asynch_position);
		gtk_label_set_text(skin->asynch_position_label, txt);
	}
}

static void do_add_asynch(GdamSequencerSkin* skin){
	GdamSequencer* sequencer = get_sequencer(skin);
	int buffer_id = gdam_buffer_manager_obtain_id(GDAM_MODEL(sequencer));
	GdamSequencerElement* sample_info;
	sample_info = gdam_sequencer_add_element(
				sequencer,
	                        buffer_id,
				skin->asynch_position,
		                skin->default_resample,
                                skin->default_gain,
	                        0,
				-1);
	if (skin->asynch_autostep){
		gdam_debug("adding asynch: autostepping.");
		do_step_asynch(skin);
	}
	delayed_redraw(skin);
}

static void do_autostep_asynch(GdamSequencerSkin* skin)
{
	gboolean rv;
	gdam_debug("autostep toggled.");
	if (!skin->autostep_toggle_button)
		return;
	rv = gtk_toggle_button_get_active(skin->autostep_toggle_button);
	skin->asynch_autostep = rv;
}

static void do_add_immediate(IntInfo* info){
	GdamSequencerSkin* skin = info->skin;
	GdamSequencer* sequencer = GDAM_SEQUENCER(GDAM_SKIN(skin)->model);
	int buffer_id;
	gboolean must_unref = FALSE;

	if (info->num && info->num < SEQUENCER_NUM_LOCAL_BUFFERS){
		buffer_id = skin->local_buffers[info->num];
	}
	else {
		buffer_id = gdam_buffer_manager_obtain_id(
						GDAM_MODEL(sequencer));
		must_unref = TRUE;
	}

	if (buffer_id == 0) {
		g_message("no sample has been chosen -- cannot add immediate");
		return;
	}

	gdam_sequencer_add_element_immediate(sequencer,
					     buffer_id,
					     skin->default_resample,
					     skin->default_gain,
					     0,
					     -1);
	if (must_unref)
		gdam_buffer_manager_unref(buffer_id, (GdamModel*)sequencer);
	delayed_redraw(skin);
}


static gboolean gdam_sequencer_drawing_area_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	if (!GTK_IS_DRAWING_AREA(object)) {
		g_warning("%s may only be connected to a drawing area",
			handler_names[0]);
	}
	connect_drawing_area(GTK_DRAWING_AREA(object),
	                     GDAM_SEQUENCER_SKIN(skin));
	return TRUE;
}

static gboolean gdam_sequencer_tempo_drop_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gdam_widget_string_dest_set(GTK_WIDGET(object),
				    on_tempo_string_drop,
				    "gdam-tempo-text",
				    skin);
	return TRUE;
}

static gboolean gdam_sequencer_sort_by_resample_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(
			object,
			signal_name,
			GTK_SIGNAL_FUNC(sort_by_resample),
			(GtkObject*) skin);
	return TRUE;
}

static gboolean gdam_sequencer_sort_by_gain_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(
			object,
			signal_name,
			GTK_SIGNAL_FUNC(sort_by_gain),
			(GtkObject*) skin);
	return TRUE;
}

static gboolean gdam_sequencer_play_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(
			object,
			signal_name,
			GTK_SIGNAL_FUNC(gdam_producer_play_now),
			(GtkObject*) GDAM_SEQUENCER(skin->model));
	return TRUE;
}

static gboolean gdam_sequencer_asynch_position_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamSequencerSkin* s_skin = (GdamSequencerSkin*)skin;
	s_skin->asynch_position_label = GTK_LABEL(object);
	return TRUE;
}

static gboolean gdam_sequencer_add_asynch_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(
			object,
			signal_name,
			GTK_SIGNAL_FUNC(do_add_asynch),
			(GtkObject*)skin);
	return TRUE;
}

static gboolean gdam_sequencer_step_asynch_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(
			object,
			signal_name,
			GTK_SIGNAL_FUNC(do_step_asynch),
			(GtkObject*)skin);
	return TRUE;
}

static gboolean gdam_sequencer_autostep_asynch_tbutton_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamSequencerSkin* seq_skin;
	seq_skin = GDAM_SEQUENCER_SKIN(skin);
	seq_skin->autostep_toggle_button = GTK_TOGGLE_BUTTON(object);
	gtk_signal_connect_object(
			object,
			signal_name,
			GTK_SIGNAL_FUNC(do_autostep_asynch),
			(GtkObject*)skin);
	return TRUE;
}

static gboolean gdam_sequencer_add_immediate_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	int num;
	IntInfo* info;
	char* endp;

	if (handler_names[1] == NULL) {
		num = 0;
	}
	else{
		num = strtod(handler_names[1], &endp);
		if (endp == handler_names[1] || *endp != '\0')
			return TRUE;
	}

	info = g_new(IntInfo, 1);
	info->skin = GDAM_SEQUENCER_SKIN(skin);
	info->num = num;

	gtk_signal_connect_object(
			object,
			signal_name,
			GTK_SIGNAL_FUNC(do_add_immediate),
			(GtkObject*)info);
	return TRUE;
}

static gboolean gdam_sequencer_select_local_buffer_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	int num;
	IntInfo* info;
	char* endp;

	if (handler_names[1] == NULL) {
		g_warning("Missing needed arg to %s", handler_names[0]);
		return TRUE;
	}
	else{
		num = strtod(handler_names[1], &endp);
		if (endp == handler_names[1] || *endp != '\0')
			return TRUE;
	}

	info = g_new(IntInfo, 1);
	info->skin = GDAM_SEQUENCER_SKIN(skin);
	info->num = num;

	gtk_signal_connect_object(
			object,
			signal_name,
			GTK_SIGNAL_FUNC(do_select_local_buffer),
			(GtkObject*)info);
	gdam_widget_string_dest_set(GTK_WIDGET(object),
				    on_buffer_id_drop,
				    "gdam-buffer-id-text",
				    (GtkObject*)info);
	return TRUE;
}

static gboolean gdam_sequencer_stop_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(
			object,
			signal_name,
			GTK_SIGNAL_FUNC(gdam_producer_stop),
			(GtkObject*) GDAM_SEQUENCER(skin->model));
	return TRUE;
}

static gboolean gdam_sequencer_set_lag_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkAdjustment* adj;
	adj = util_get_widget_adjustment(GTK_WIDGET(object));
	gtk_signal_connect(
			GTK_OBJECT(adj),
			"value_changed",
			GTK_SIGNAL_FUNC(on_lag_adjusted),
			skin);
	gdam_delayed_adjustment_init(adj);
	return TRUE;
}

static gboolean gdam_sequencer_set_asynch_step_size_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkAdjustment* adj;
	adj = util_get_widget_adjustment(GTK_WIDGET(object));
	gtk_signal_connect(
			GTK_OBJECT(adj),
			"value_changed",
			GTK_SIGNAL_FUNC(on_asynch_step_size_adjusted),
			skin);
	gdam_delayed_adjustment_init(adj);
	return TRUE;
}

static gboolean gdam_sequencer_set_sub_beats_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkAdjustment* adj;
	adj = util_get_widget_adjustment(GTK_WIDGET(object));
	gtk_signal_connect(
			GTK_OBJECT(adj),
			"value_changed",
			GTK_SIGNAL_FUNC(on_subbeats_adjusted),
			skin);
	gdam_delayed_adjustment_init(adj);
	return TRUE;
}

static gboolean gdam_sequencer_beatnum_adjuster_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkAdjustment* adj;
	adj = util_get_widget_adjustment(GTK_WIDGET(object));
	gtk_signal_connect(
			GTK_OBJECT(adj),
			"value_changed",
			GTK_SIGNAL_FUNC(on_beatnum_adjusted),
			skin);
	gdam_delayed_adjustment_init(adj);
	return TRUE;
}

static gboolean gdam_sequencer_tempo_adjuster_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamSequencerSkin* seq_skin = GDAM_SEQUENCER_SKIN(skin);
	GtkAdjustment* adj;
	adj = util_get_widget_adjustment(GTK_WIDGET(object));
	gtk_signal_connect(
			GTK_OBJECT(adj),
			"value_changed",
			GTK_SIGNAL_FUNC(on_tempo_adjusted),
			skin);
	gtk_signal_connect_object(
			GTK_OBJECT(skin->model),
			"set_beat_info",
			GTK_SIGNAL_FUNC(on_sequencer_set_beat_info),
			GTK_OBJECT(adj));
	seq_skin->tempo_adj = adj;
	gdam_delayed_adjustment_init(adj);
	return TRUE;
}

static gboolean gdam_sequencer_h_zoom_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkAdjustment* adj;
	GdamSequencerSkin* sequencer_skin = GDAM_SEQUENCER_SKIN(skin);
	adj = util_get_widget_adjustment(GTK_WIDGET(object));
	sequencer_skin->time_zoom_adj = adj;
	gtk_signal_connect(GTK_OBJECT(adj),
			   "value_changed",
			   GTK_SIGNAL_FUNC(horizontal_zoom),
			   skin);
	gdam_delayed_adjustment_init(adj);
	return TRUE;
}

static gboolean gdam_sequencer_h_pan_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkAdjustment* adj;
	GdamSequencerSkin* sequencer_skin = GDAM_SEQUENCER_SKIN(skin);
	adj = util_get_widget_adjustment(GTK_WIDGET(object));
	sequencer_skin->time_pan_adj = adj;
	gtk_signal_connect(GTK_OBJECT(adj),
			   "value_changed",
			   GTK_SIGNAL_FUNC(horizontal_pan),
			   skin);
	gdam_delayed_adjustment_init(adj);
	return TRUE;
}

static gboolean gdam_sequencer_v_zoom_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkAdjustment* adj;
	GdamSequencerSkin* sequencer_skin = GDAM_SEQUENCER_SKIN(skin);
	adj = util_get_widget_adjustment(GTK_WIDGET(object));
	sequencer_skin->vertical_zoom_adj = adj;
	gtk_signal_connect(GTK_OBJECT(adj),
			   "value_changed",
			   GTK_SIGNAL_FUNC(vertical_zoom),
			   skin);
	gdam_delayed_adjustment_init(adj);
	return TRUE;
}

static gboolean gdam_sequencer_v_pan_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkAdjustment* adj;
	GdamSequencerSkin* sequencer_skin = GDAM_SEQUENCER_SKIN(skin);
	adj = util_get_widget_adjustment(GTK_WIDGET(object));
	sequencer_skin->vertical_pan_adj = adj;
	gtk_signal_connect(GTK_OBJECT(adj),
			   "value_changed",
			   GTK_SIGNAL_FUNC(vertical_pan),
			   skin);
	gdam_delayed_adjustment_init(adj);
	return TRUE;
}

static gboolean gdam_sequencer_filename_entry_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamSequencerSkin* sequencer_skin = GDAM_SEQUENCER_SKIN(skin);
	sequencer_skin->filename_entry = GTK_ENTRY(object);
	return TRUE;
}

static gboolean gdam_sequencer_midi_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(
			object,
			signal_name,
			GTK_SIGNAL_FUNC(do_midi),
			GTK_OBJECT(skin));
	return TRUE;
}

static gboolean gdam_sequencer_save_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(
			object,
			signal_name,
			GTK_SIGNAL_FUNC(do_save),
			GTK_OBJECT(skin));
	return TRUE;
}

static gboolean gdam_sequencer_load_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(
			object,
			signal_name,
			GTK_SIGNAL_FUNC(do_load),
			GTK_OBJECT(skin));
	return TRUE;
}

static gboolean gdam_sequencer_buffer_container_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	if (! GTK_IS_CONTAINER(object) ) {
		g_warning("arg to %s must be a container", handler_names[0]);
		return TRUE;
	}
	if (GTK_IS_BIN(object) ) {
		g_warning("arg to %s must not be a bin", handler_names[0]);
		return TRUE;
	}
	GDAM_SEQUENCER_SKIN(skin)->buffer_id_holder = GTK_CONTAINER(object);
	return TRUE;
}

static void recompute_beat_grid(GdamSequencerSkin* skin)
{
	int i;
	gdouble bpm = gdam_beat_info_get_bpm(
				&GDAM_MODEL(get_sequencer(skin))->beat_info,
				0);
	gdouble one_grid_length;
	GdamSequencer* sequencer = get_sequencer(skin);
	if (bpm <= 0.0) 
		return;
	if (skin->beat_grid)
		g_free(skin->beat_grid);
	skin->num_beat_grid = sequencer->num_beats * sequencer->num_sub_beats;
	if (sequencer->num_sub_beats == 0)
		one_grid_length = 1.0;
	else
		one_grid_length = 1.0 / sequencer->num_sub_beats;
	skin->beat_grid = g_new(gdouble, skin->num_beat_grid);
	for (i = 0; i < skin->num_beat_grid; i++)
		skin->beat_grid[i] = one_grid_length * i;
	gdam_sequencer_set_num_beats(sequencer, sequencer->num_beats);
}

static int redraw_skin_now(gpointer skin_ptr)
{
	GdamSequencerSkin* skin = GDAM_SEQUENCER_SKIN(skin_ptr);
	g_assert(skin->redraw_thread != 0);
	skin->redraw_thread = 0;
	redraw_area(skin, 0, 0, skin->draw_width, skin->draw_height);
	return 0;
}

static void delayed_redraw(GdamSequencerSkin* skin)
{
	if (skin->redraw_thread)
		return;
	skin->redraw_thread = g_timeout_add(100,
	                                    redraw_skin_now,
					    skin);
}

static void on_lag_adjusted(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
	float new_lag = adj->value;
	GdamSequencer* sequencer = get_sequencer(skin);
	sequencer->immediate_lag = new_lag;
}

static void on_asynch_step_size_adjusted(GtkAdjustment* adj, 
					 GdamSequencerSkin* skin)
{
	skin->asynch_step_size = 1/adj->value;
}

static void on_subbeats_adjusted(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
	int new_subbeats = adj->value;
	gboolean redraw = FALSE;
	GdamSequencer* sequencer = get_sequencer(skin);
	if ((new_subbeats == 0 && sequencer->num_sub_beats != 0)
	 || (new_subbeats != 0 && sequencer->num_sub_beats == 0))
	 	redraw = TRUE;
	/* TODO:  signal to sequencer?  Recompute sequence here? */
	sequencer->num_sub_beats = new_subbeats;
	if (redraw)
		delayed_redraw(skin);
	recompute_beat_grid(skin);
}

static void on_beatnum_adjusted(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
	gdam_debug("beatnum_adjusted");
	/* TODO:  signal to sequencer?  Recompute sequence here? */
	get_sequencer(skin)->num_beats = adj->value;
	skin->min_beat = 0.0;
	skin->max_beat = adj->value;
	if (skin->time_zoom_adj) {
		skin->time_zoom_adj->value = adj->value;
		skin->time_zoom_adj->upper = adj->value;
	}
	if (skin->time_pan_adj) {
		skin->time_pan_adj->value = 0.0;
		skin->time_pan_adj->upper = skin->time_pan_adj->page_size
		                           + adj->value;
	}
	if (skin->time_zoom_adj)
		gtk_adjustment_changed(skin->time_zoom_adj);
	if (skin->time_pan_adj)
		gtk_adjustment_changed(skin->time_pan_adj);
	recompute_beat_grid(skin);
	delayed_redraw(skin);
}

static void on_tempo_adjusted(GtkAdjustment* adj, gpointer user_data)
{
	GdamSequencerSkin* skin = (GdamSequencerSkin*)user_data;
	GdamBeatInfo new_beat_info;
	new_beat_info = GDAM_MODEL(get_sequencer(skin))->beat_info;
	gdam_beat_info_set_bpm(&new_beat_info, adj->value);
	//new_beat_info.bpm = adj->value;
	gdam_model_set_beat_info(GDAM_MODEL(get_sequencer(skin)), 
						    &new_beat_info);
	recompute_beat_grid(skin);
	delayed_redraw(skin);
}

static void on_sequencer_set_beat_info(GtkAdjustment* adj, 
				       const GdamBeatInfo* info, 
				       gpointer user_data)
{
	gtk_adjustment_set_value(adj, gdam_beat_info_get_bpm(info,0)); 
}

static void on_buffer_id_drop(const char* id_string, gpointer data)
{
	IntInfo* info = (IntInfo*)data;
	GdamSequencerSkin* skin = info->skin;
	//GdamModel* component = GDAM_SKIN(skin)->component;
	int id = (int)atof(id_string);
	int index = info->num;
	if (index < SEQUENCER_NUM_LOCAL_BUFFERS){
		skin->local_buffers[index] = id;
		gdam_debug("setting local buffer %d to %d", index, id);
	}
	else{
		g_warning("error, only buffers 0 through %d allowed.",
						SEQUENCER_NUM_LOCAL_BUFFERS-1);
	}
}

static void on_tempo_string_drop(const char* tempo, gpointer user_data)
{
	GdamSequencerSkin* skin = (GdamSequencerSkin*)user_data;
	GdamModel* model = GDAM_SKIN(skin)->model;
	GdamBeatInfo new_beat_info = model->beat_info;
	gdam_beat_info_set_bpm(&new_beat_info,  atof(tempo));
	gdam_model_set_beat_info(model, &new_beat_info);

}

/* TODO: add this. */
static void on_set_bi(GdamProducer* 		producer,
		      const GdamBeatInfo* 	bi,
		      GdamSequencerSkin* 	skin)
{
	if (skin->tempo_adj){
		//skin->tempo_adj->value = gdam_beat_info_get_bpm(bi,0);
		//gtk_adjustment_value_changed(skin->tempo_adj);
	}
}

static void on_sequence_changed(GdamSequencer* sequencer,
                                GdamSequencerSkin* skin)
{
	delayed_redraw(skin);
}

static gboolean       gdam_sequencer_build_new_model
                                                 (GdamSkin*      skin,
                                                  GdamChannel*   channel,
						  int		 num_options,
						  char**         options)
{
	GdamModel* model;
	model = GDAM_MODEL(gtk_type_new(GDAM_TYPE_SEQUENCER));
	model->channel = channel;
	skin->model = model;
	gtk_signal_connect(GTK_OBJECT(model),
			   "set_beat_info",
			   on_set_bi,
			   skin);
	gtk_signal_connect(GTK_OBJECT(model),
			   "sequence_changed",
			   on_sequence_changed,
			   skin);
	return TRUE;
}
static GdamSkinGenericConnect connect_infos[] =
{
	{ "gdam_sequencer_drawing_area",
           gdam_sequencer_drawing_area_handler			},
	{ "gdam_sequencer_tempo_drop",
           gdam_sequencer_tempo_drop_handler			},
	{ "gdam_sequencer_sort_by_resample",
           gdam_sequencer_sort_by_resample_handler		},
	{ "gdam_sequencer_sort_by_gain",
           gdam_sequencer_sort_by_gain_handler			},
	{ "gdam_sequencer_play",
           gdam_sequencer_play_handler				},
	{ "gdam_sequencer_stop",
           gdam_sequencer_stop_handler				},
	{ "gdam_sequencer_add_immediate",
           gdam_sequencer_add_immediate_handler			},
	{ "gdam_sequencer_add_asynch",
           gdam_sequencer_add_asynch_handler			},
	{ "gdam_sequencer_asynch_position_label",
           gdam_sequencer_asynch_position_label_handler		},
	{ "gdam_sequencer_step_asynch",
           gdam_sequencer_step_asynch_handler			},
	{ "gdam_sequencer_autostep_asynch_tbutton",
           gdam_sequencer_autostep_asynch_tbutton_handler	},
	{ "gdam_sequencer_select_local_buffer",
           gdam_sequencer_select_local_buffer_handler		},
	{ "gdam_sequencer_set_sub_beats",
           gdam_sequencer_set_sub_beats_handler			},
	{ "gdam_sequencer_set_asynch_step_size",
           gdam_sequencer_set_asynch_step_size_handler		},
	{ "gdam_sequencer_set_lag",
           gdam_sequencer_set_lag_handler			},
	{ "gdam_sequencer_beatnum_adjuster",
           gdam_sequencer_beatnum_adjuster_handler		},
	{ "gdam_sequencer_tempo_adjuster",
           gdam_sequencer_tempo_adjuster_handler		},
	{ "gdam_sequencer_h_zoom",
           gdam_sequencer_h_zoom_handler			},
	{ "gdam_sequencer_h_pan",
           gdam_sequencer_h_pan_handler				},
	{ "gdam_sequencer_v_zoom",
           gdam_sequencer_v_zoom_handler			},
	{ "gdam_sequencer_v_pan",
           gdam_sequencer_v_pan_handler				},
	{ "gdam_sequencer_filename_entry",
           gdam_sequencer_filename_entry_handler		},
	{ "gdam_sequencer_midi",
           gdam_sequencer_midi_handler				},
	{ "gdam_sequencer_save",
           gdam_sequencer_save_handler				},
	{ "gdam_sequencer_load",
           gdam_sequencer_load_handler				},
	{ "gdam_sequencer_buffer_container",
           gdam_sequencer_buffer_container_handler		}
};

static void gdam_sequencer_skin_destroy(GtkObject* object)
{
	GdamSequencerSkin* skin = GDAM_SEQUENCER_SKIN(object);
	if (skin->selected_elements != NULL) {
		g_hash_table_destroy(skin->selected_elements);
		skin->selected_elements = NULL;
	}
	if (skin->grid_gc != NULL) {
		gdk_gc_unref(skin->grid_gc);
		skin->grid_gc = NULL;
	}
	if (skin->circle_boundary_gc != NULL) {
		gdk_gc_unref(skin->circle_boundary_gc);
		skin->circle_boundary_gc = NULL;
	}
	if (skin->backing_pixmap != NULL) {
		gdk_pixmap_unref(skin->backing_pixmap);
		skin->backing_pixmap = NULL;
	}
	if (skin->in_transit != NULL) {
		g_hash_table_destroy(skin->in_transit);
		skin->in_transit = NULL;
	}

	(*parent_class->destroy)(object);
}

static void gdam_sequencer_skin_class_init(GdamSkinClass* skin_class)
{
	skin_class->build_new_model = gdam_sequencer_build_new_model;
	GTK_OBJECT_CLASS(skin_class)->destroy = gdam_sequencer_skin_destroy;
	gdam_skin_class_register_generic_handlers(skin_class,
				                  G_N_ELEMENTS(connect_infos),
				                  connect_infos);
}

